"""Base infrastructure of regelum."""

import abc

from regelum import callback as cb

import weakref


class apply_callbacks:
    """Decorator that applies a list of callbacks to a given method of an object.

    If the object has no list of callbacks specified, the default callbacks are used.

    Args:
        method: The method to which the callbacks should be applied.
    """

    def __init__(self, callbacks=None):
        """Initialize a decorator that applies callbacks.

        Args:
            callbacks: list of callbacks to apply (applies all default
                callbacks if omitted)
        """
        self.callbacks = callbacks

    def __call__(self, method):
        def new_method(self2, *args, **kwargs):
            res = method(self2, *args, **kwargs)
            if self.callbacks is None:
                try:
                    callbacks = RegelumBase._metadata["main"].callbacks
                except AttributeError:
                    callbacks = []
            for callback in callbacks:
                callback(obj=self2, method=method.__name__, output=res)
            return res

        return new_method


class RegelumType(abc.ABCMeta):
    """Regelum type that all classes in regelum share.

    Used for certain infrastructural and syntactic sugar features.
    """

    def __str__(self):
        return f"<class '{self.__name__}'>"

    def __repr__(self):
        return str(self)

    @classmethod
    def __register_callback(cls, callback):
        pass

    @classmethod
    def __prepare__(metacls, *args):
        return {"apply_callbacks": apply_callbacks} | super().__prepare__(*args)

    def __add__(self, other):
        assert hasattr(self, "_compose")
        return self._compose(other)

    def __radd__(self, other):
        if other == 0:
            return self

    def __new__(cls, *args, **kwargs):
        x = super().__new__(cls, *args, **kwargs)
        if (
            x.__name__ == "RegelumBase"
            or x.__name__ == "Callback"
            or issubclass(x, cb.Callback)
        ):
            return x
        if hasattr(x, "_real_name"):
            x.__name__ = x._real_name
            del x._real_name
        # callbacks = x._attached if hasattr(x, "_attached") else []
        # if callbacks:
        #    del x._attached
        x._callbacks_registered = False

        if x.__init__ is not x.__bases__[0].__init__:
            original_init = x.__init__

            def pre_init(
                self, *args, _tracked=True, original_init=original_init, **kwargs
            ):
                if not self._callbacks_registered and _tracked:
                    callbacks = self._attached if hasattr(self, "_attached") else []
                    animations = [
                        callback
                        for callback in callbacks
                        if (
                            issubclass(callback, cb.AnimationCallback)
                            and (not callback.is_jupyter)
                        )
                    ]
                    non_animations = [
                        callback
                        for callback in callbacks
                        if not issubclass(callback, cb.AnimationCallback)
                    ]
                    for callback in non_animations:
                        callback.register(attachee=x, launch=True)
                    if animations and not self._metadata['argv'].parallel:
                        if animations:
                            composed = sum(animations) + None
                            composed.register(attachee=x, launch=True)
                self._callbacks_registered = True
                return original_init(self, *args, **kwargs)

            x.__init__ = pre_init

        return x


class ClassPropertyDescriptor(object):
    """Enables to declare class properties."""

    def __init__(self, fget, fset=None):
        """Initialize an instance of ClassPropertyDescriptor.

        Args:
            fget: class getter
            fset: class setter
        """
        self.fget = fget
        self.fset = fset

    def __get__(self, obj, cls=None):
        if cls is None:
            cls = type(obj)
        return self.fget.__get__(obj, cls)()

    def __set__(self, obj, value):
        if not self.fset:
            raise AttributeError("can't set attribute")
        type_ = type(obj)
        return self.fset.__get__(obj, type_)(value)

    def setter(self, func):
        if not isinstance(func, (classmethod, staticmethod)):
            func = classmethod(func)
        self.fset = func
        return self


def classproperty(func):
    """Decorate a method in such a way that it becomes a class property.

    Args:
        func: method to decorate

    Returns:

    """
    if not isinstance(func, (classmethod, staticmethod)):
        func = classmethod(func)

    return ClassPropertyDescriptor(func)


class RegelumBase(metaclass=RegelumType):
    """Base class designed to act as an abstraction over all regelum objects."""

    @classproperty
    def _metadata(self):
        return RegelumBase.__metadata

    @_metadata.setter
    def _metadata(self, metadata):
        if not hasattr(RegelumBase, f"_{RegelumBase.__name__}__metadata"):
            RegelumBase.__metadata = metadata
        else:
            raise ValueError(
                "Metadata has already been set, yet an attempt to set it again was made."
            )

    def __init__(self):
        """Initialize an object from regelum."""
        # if hasattr(self.__class__, "_attached"):
        #    existing_callbacks = [type(callback) for callback in regelum.main.callbacks]
        #    for callback in self._attached:
        #        if callback not in existing_callbacks:
        #            callback_instance = callback(attachee=self.__class__)  ## Might want to move it to the metaclass
        #            callback_instance.on_launch()  # I must change this
        #            regelum.main.callbacks = [callback_instance] + regelum.main.callbacks

        # callbacks = self.__class__._attached if hasattr(self.__class__, "_attached") else []
        # existing_callbacks = [type(callback) for callback in regelum.main.callbacks]
        # for callback in callbacks:
        #    if callback not in existing_callbacks:
        #        callback_instance = callback(attachee=self.__class__) ## Might want to move it to the metaclass
        #        callback_instance.on_launch()
        #        regelum.main.callbacks = [callback_instance] + regelum.main.callbacks


class Node(abc.ABC):
    """A node is an object that is responsible for sending/receiving/rerouting/processing messages."""

    def __init__(self, input_type):
        """Initialize a node.

        Args:
            input_type (type): type of messages sent/received by the
                node
        """
        self.__subscribers = []
        self.__subscribees = []
        self.hooks = []
        self.type = input_type

    # TODO: DOCSTRING
    def hook(self, hook_function):
        def new_hook_function(input_):
            res = hook_function(input_)
            assert isinstance(
                res, self.type
            ), f"Values returned by hooks should match the type of their node (port/publisher). An object of type {type(res)} was returned, which does not match the node's type {self.type}."
            return res

        self.hooks.append(new_hook_function)

    # TODO: DOCSTRING
    def __forget(self, other):
        assert self.connected(
            other
        ), "Attempt to disconnect a node that was not connected."
        for i, subscriber_ref in enumerate(self.__subscribers):
            subscriber = subscriber_ref()
            if subscriber is other:
                self.__subscribers.pop(i)
                break
        for i, subscribee_ref in enumerate(self.__subscribees):
            subscribee = subscribee_ref()
            if subscribee is other:
                self.__subscribees.pop(i)
                break

    # TODO: DOCSTRING
    def connected(self, other):
        return other in self.subscribees or other in self.subscribers

    # TODO: DOCSTRING
    def disconnect(self, other):
        self.__forget(other)
        other.__forget(self)

    # TODO: DOCSTRING
    def __del__(self):
        for subscriber in self.subscribers:
            self.disconnect(subscriber)
        for subscribee in self.subscribees:
            self.disconnect(subscribee)

    # TODO: DOCSTRING
    def __subscribe(self, other):
        assert isinstance(
            other, Node
        ), "Attempt to subscribe to something that is neither a Port nor a Publisher."
        assert issubclass(
            other.type, self.type
        ), f"Type mismatch. Attempt to subscribe a node of type {self.type} to node of type {other.type}."
        self.__subscribees.append(weakref.ref(other))
        other.__subscribers.append(weakref.ref(self))

    # TODO: DOCSTRING
    def __issue_subscription(self, other):
        assert isinstance(
            other, Node
        ), "Attempt to issue subscription to something that is neither a Port nor a Publisher."
        assert issubclass(
            self.type, other.type
        ), f"Type mismatch. Attempt to subscribe a node of type {other.type} to node of type {self.type}."
        self.__subscribers.append(weakref.ref(other))
        other.__subscribees.append(weakref.ref(self))

    @property
    def subscribers(self):
        return [
            subscriber()
            for subscriber in self.__subscribers
            if subscriber() is not None
        ]

    @property
    def subscribees(self):
        return [
            subscribee()
            for subscribee in self.__subscribees
            if subscribee() is not None
        ]

    @abc.abstractmethod
    def connect(self, other):
        pass

    @abc.abstractmethod
    def __on_input(self, message):
        pass

    def __call__(self, message):
        assert isinstance(
            message, self.type
        ), f"Type mismatch. Attempt to pass a message of type {type(message)} to node of type {self.type}."
        for hook in self.hooks:
            message = hook(message)
        return self.__on_input(message)


# TODO: DOCSTRING
class EmptyInboxException(Exception):
    """Raised when trying to receive a message from a Port, to which no messages were yet sent."""

    pass


# TODO: DOCSTRING
class port:
    """Decorator factory that replaces a method with a Port in such a way that the initial definition of the method is interpreted as said Port's handler."""

    def __init__(self, input_type=object, hooks=None):
        """Initialize a decorator that transforms methods into handled ports.

        Args:
            input_type (type): type of messages accepted by the
                resulting port
            hooks: hooks to add in addition to the handler obtained from
                the decorated method
        """
        if hooks is None:
            hooks = []
        self.input_type = input_type
        self.hooks = hooks

    def __call__(self, method):
        new_port = Port(self.input_type)
        new_port.handle(method)
        for hook in self.hooks:
            new_port.hook(hook)
        return new_port


class publisher:
    """Decorator factory that replaces a method with a Publisher in such a way that the initial definition of the method is interpreted as said Publisher's hook."""

    def __init__(self, input_type=object, hooks=None):
        """Initialize a decorator that transforms methods into hooked publishers.

        Args:
            input_type (type): type of messages sent by the resulting
                publisher
            hooks: hooks to add in addition to the hook obtained from
                the decorated method
        """
        if hooks is None:
            hooks = []
        self.input_type = input_type
        self.hooks = hooks

    def __call__(self, method):
        new_port = Publisher(self.input_type)
        new_port.hook(method)
        for hook in self.hooks:
            new_port.hook(hook)
        return new_port


class Port(Node):
    """A Port is a Node that accepts messages."""

    def __init__(self, input_type):
        """Initialize a Port.

        Args:
            input_type (type): type of messages received by the port.
        """
        super().__init__(input_type)
        self.__inbox = []
        self.__handler = None

    def connect(self, other):
        self.__subscribe(other)

    def __on_input(self, message):
        for subscriber in self.subscribers:
            subscriber(message)
        if self.__handler is None:
            self.__inbox.append(message)
        else:
            self.__handler(message)

    def handle(self, handler):
        self.__handler = handler

    def receive(self):
        assert self.handler is None, "``receive`` called on a handled port."
        if self.__inbox:
            return self.__inbox.pop(-1)
        else:
            raise EmptyInboxException(
                "The port's inbox was empty at the time of calling ``receive``."
            )


# TODO: DOCSTRING
class Publisher(Node):
    """A Publisher is a Node that sends messages."""

    def connect(self, other):
        self.__issue_subscription(other)

    def __on_input(self, message):
        for subscriber in self.subscribers:
            subscriber(message)


class FreePort(Port):
    """A Port that can accept messages of arbitrary type."""

    def __init__(self):
        """Initialize a free port."""
        super().__init__(object)


class FreePublisher(Publisher):
    """A Publisher that can send messages of arbitrary type."""

    def __init__(self):
        """Initialize a free publisher."""
        super().__init__(object)


"""
class LazyMessageNotYetReceivedException(Exception):
    pass
class LazyMessage:
    def __init__(self, container, index, message_type):
        self.type = message_type
        self.container = container
        self.index = index

    def __str__(self):
        try:
            content = str(self())
        except LazyMessageNotYetReceivedException:
            content = "N/A"
        return f"{self.__class__.__name__}({content})"

    def __repr__(self):
        try:
            content = repr(self())
        except LazyMessageNotYetReceivedException:
            content = "N/A"
        return f"{self.__class__.__name__}({content})"

    def __call__(self):
        if len(self.container) < self.index:
            return self.container[self.index]
        else:
            raise LazyMessageNotYetReceivedException("Attempted to unbox a lazy message, but the message was not yet received.")

    def __add__(self, other):
        if isinstance(other, LazyMessage):


class LazyPort(Port):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.index = 0

    def receive(self):
        assert self.handler is None, "``receive`` called on a handled port."
        message = LazyMessage(container=self.__inbox, index=self.index, message_type=self.type)
        self.index += 1
        return message
"""



"""Contains high-level structures of scenarios (agents).

Remarks:

- All vectors are treated as of type [n,]
- All buffers are treated as of type [L, n] where each row is a vector
- Buffers are updated from bottom to top

"""

import numpy as np
import torch
import torch.multiprocessing as mp
import random

from regelum.data_buffers import DataBuffer

from regelum.utils import Clock, AwaitedParameter, calculate_value
from regelum import RegelumBase
from .policy import Policy, RLPolicy, PolicyPPO, PolicyReinforce, PolicySDPG, PolicyDDPG
from .critic import Critic, CriticCALF, CriticTrivial
from typing import Optional, Union, Type, Dict, List, Any, Callable
from .objective import RunningObjective
from .data_buffers import DataBuffer
from .event import Event
from . import OptStatus
from .simulator import Simulator
from .constraint_parser import ConstraintParser, ConstraintParserTrivial
from .observer import Observer, ObserverTrivial
from .model import (
    Model,
    ModelNN,
    ModelPerceptron,
    ModelWeightContainer,
    ModelWeightContainerTorch,
    PerceptronWithTruncatedNormalNoise,
    ModelQuadLin,
)
from .predictor import Predictor, EulerPredictor
from regelum.optimizable.core.configs import (
    TorchOptimizerConfig,
    CasadiOptimizerConfig,
)
from regelum.data_buffers.batch_sampler import RollingBatchSampler, EpisodicSampler
from copy import deepcopy
from typing_extensions import Self


class Scenario(RegelumBase):
    """Scenario orchestrator.

    A Scenario orchestrates the training and evaluation cycle of a reinforcement learning agent.
    It runs the simulation based on a given policy, collects observations, applies actions, and
    manages the overall simulation loop, including assessing the agent's performance.
    """

    def __init__(
        self,
        policy: Policy,
        simulator: Simulator,
        sampling_time: float = 0.1,
        running_objective: Optional[RunningObjective] = None,
        constraint_parser: Optional[ConstraintParser] = None,
        observer: Optional[Observer] = None,
        N_episodes: int = 1,
        N_iterations: int = 1,
        value_threshold: float = np.inf,
        discount_factor: float = 1.0,
    ):
        """Initialize the Scenario with the necessary components for running a reinforcement learning experiment.

        Args:
            policy: Policy to generate actions based on observations.
            simulator: Simulator to interact with and collect data for
                training.
            sampling_time: Time interval between action updates.
            running_objective: Objective function for evaluating
                performance.
            constraint_parser: Tool for parsing constraints during
                policy optimization.
            observer: Observer for estimating the system state.
            N_episodes: Total number of episodes to run.
            N_iterations: Total number of iterations to run.
            value_threshold: Threshold to stop the simulation if the
                objective is met.
            discount_factor: Discount factor for future rewards.
        """
        super().__init__()
        self.N_episodes = N_episodes
        self.N_iterations = N_iterations
        self.simulator = simulator
        self.time_old = 0
        self.delta_time = 0
        self.value: float = 0.0
        self.recent_values_of_episodes = []

        self.sim_status = 1
        self.episode_counter = 0
        self.iteration_counter = 0
        self.value_threshold = value_threshold
        self.discount_factor = discount_factor
        self.is_episode_ended = False
        self.constraint_parser = (
            ConstraintParserTrivial()
            if constraint_parser is None
            else constraint_parser
        )
        self.observer = observer if observer is not None else ObserverTrivial()

        self.state_init, self.action_init = AwaitedParameter(
            "state_init", awaited_from=self.simulator.__class__.__name__
        ), AwaitedParameter(
            "action_init", awaited_from=self.simulator.__class__.__name__
        )
        self.state = self.state_init
        self.action = self.action_init
        self.observation = AwaitedParameter(
            "observation", awaited_from=self.simulator.system.get_observation.__name__
        )

        self.policy = policy
        self.sampling_time = sampling_time
        self.clock = Clock(period=sampling_time)
        self.iteration_counter: int = 0
        self.episode_counter: int = 0
        self.step_counter: int = 0
        self.action_old = AwaitedParameter(
            "action_old", awaited_from=self.compute_action.__name__
        )
        self.running_objective = (
            running_objective
            if running_objective is not None
            else lambda *args, **kwargs: 0
        )
        self.reset()

    def run(self):
        for iteration_counter in range(1, self.N_iterations + 1):
            for episode_counter in range(1, self.N_episodes + 1):
                self.run_episode(
                    episode_counter=episode_counter, iteration_counter=iteration_counter
                )
                self.reload_scenario()

            self.reset_iteration()
            if self.sim_status == "simulation_ended":
                break

    def get_action_from_policy(self):
        return self.simulator.system.apply_action_bounds(self.policy.action)

    def run_episode(self, episode_counter, iteration_counter):
        self.episode_counter = episode_counter
        self.iteration_counter = iteration_counter
        while self.sim_status != "episode_ended":
            self.sim_status = self.step()

    def step(self):
        if isinstance(self.action_init, AwaitedParameter) and isinstance(
            self.state_init, AwaitedParameter
        ):
            (
                self.state_init,
                self.action_init,
            ) = self.simulator.get_init_state_and_action()

        if (not self.is_episode_ended) and (self.value <= self.value_threshold):
            (
                self.time,
                self.state,
                self.observation,
                self.simulation_metadata,
            ) = self.simulator.get_sim_step_data()

            self.delta_time = (
                self.time - self.time_old
                if self.time_old is not None and self.time is not None
                else 0
            )
            self.time_old = self.time
            if len(list(self.constraint_parser)) > 0:
                self.constraint_parameters = self.constraint_parser.parse_constraints(
                    simulation_metadata=self.simulation_metadata
                )
                self.substitute_constraint_parameters(**self.constraint_parameters)
            estimated_state = self.observer.get_state_estimation(
                self.time, self.observation, self.action
            )

            self.action = self.compute_action_sampled(
                self.time,
                estimated_state,
                self.observation,
            )
            self.simulator.receive_action(self.action)
            self.is_episode_ended = self.simulator.do_sim_step() == -1
            return "episode_continues"
        else:
            return "episode_ended"

    @apply_callbacks()
    def reset_iteration(self):
        pass

    @apply_callbacks()
    def reload_scenario(self):
        self.is_episode_ended = False
        self.recent_value = self.value
        self.observation = self.simulator.observation
        self.sim_status = 1
        self.time = 0
        self.time_old = 0
        self.action = self.action_init
        self.simulator.reset()
        self.reset()
        self.sim_status = 0
        return self.recent_value

    @apply_callbacks()
    def post_compute_action(self, observation, estimated_state):
        return {
            "estimated_state": estimated_state,
            "observation": observation,
            "time": self.time,
            "episode_id": self.episode_counter,
            "iteration_id": self.iteration_counter,
            "step_id": self.step_counter,
            "action": self.get_action_from_policy(),
            "running_objective": self.current_running_objective,
            "current_value": self.value,
        }

    def compute_action_sampled(self, time, estimated_state, observation):
        self.is_time_for_new_sample = self.clock.check_time(time)
        if self.is_time_for_new_sample:
            self.on_observation_received(time, estimated_state, observation)
            action = self.simulator.system.apply_action_bounds(
                self.compute_action(
                    time=time,
                    estimated_state=estimated_state,
                    observation=observation,
                )
            )
            self.post_compute_action(observation, estimated_state)
            self.step_counter += 1
            self.action_old = action
        else:
            action = self.action_old
        return action

    def compute_action(self, time, estimated_state, observation):
        self.issue_action(observation)
        return self.get_action_from_policy()

    def issue_action(self, observation):
        self.policy.update_action(observation)

    def __getattribute__(self, name):
        if name == "issue_action":
            return self._issue_action
        else:
            return object.__getattribute__(self, name)

    def _issue_action(self, observation, *args, **kwargs):
        object.__getattribute__(self, "issue_action")(observation, *args, **kwargs)
        self.on_action_issued(observation)

    def on_action_issued(self, observation):
        self.current_running_objective = self.running_objective(
            observation, self.get_action_from_policy()
        )
        self.value = self.calculate_value(self.current_running_objective, self.time)
        observation_action = np.concatenate(
            (observation, self.get_action_from_policy()), axis=1
        )
        return {
            "action": self.get_action_from_policy(),
            "running_objective": self.current_running_objective,
            "current_value": self.value,
            "observation_action": observation_action,
        }

    def on_observation_received(self, time, estimated_state, observation):
        self.time = time
        return {
            "estimated_state": estimated_state,
            "observation": observation,
            "time": time,
            "episode_id": self.episode_counter,
            "iteration_id": self.iteration_counter,
            "step_id": self.step_counter,
        }

    def substitute_constraint_parameters(self, **kwargs):
        self.policy.substitute_parameters(**kwargs)

    def calculate_value(self, running_objective: float, time: float):
        value = (
            self.value
            + running_objective * self.discount_factor**time * self.sampling_time
        )
        return value

    def reset(self):
        """Reset agent for use in multi-episode simulation.

        Only __internal clock and current actions are reset.
        All the learned parameters are retained.

        """
        self.clock.reset()
        self.value = 0.0
        self.is_first_compute_action_call = True


class RLScenario(Scenario):
    """Incorporates reinforcement learning algorithms.

    The RLScenario incorporates reinforcement learning algorithms into the Scenario framework,
    enabling iterative optimization of both policies and value functions as part of the agent's learning process.
    """

    def __init__(
        self,
        policy: Policy,
        critic: Critic,
        running_objective: RunningObjective,
        simulator: Simulator,
        policy_optimization_event: Event,
        critic_optimization_event: Event = None,
        discount_factor: float = 1.0,
        is_critic_first: bool = False,
        sampling_time: float = 0.1,
        constraint_parser: Optional[ConstraintParser] = None,
        observer: Optional[Observer] = None,
        N_episodes: int = 1,
        N_iterations: int = 1,
        value_threshold: float = np.inf,
        stopping_criterion: Optional[Callable[[DataBuffer], bool]] = None,
    ):
        """Instantiate a RLScenario object.

        Args:
            policy (Policy): Policy object
            critic (Critic): Cricic
            running_objective (RunningObjective): Function to calculate
                the running objective.
            critic_optimization_event (str): moments when to optimize
                critic. Can be either 'compute_action' for online
                learning, or 'reset_episode' for optimizing after each
                episode, or 'reset_iteration' for optimizing after each
                iteration
            policy_optimization_event (str): moments when to optimize
                critic. Can be either 'compute_action' for online
                learning, or 'reset_episode' for optimizing after each
                episode, or 'reset_iteration' for optimizing after each
                iteration
            discount_factor (float, optional): Discount factor. Used for
                computing value as discounted sum (or
                integral) of running objectives, defaults to 1.0
            is_critic_first (bool, optional): if is True then critic is
                optimized first then policy (can be usefull in DQN or
                Predictive Algorithms such as RPV, RQL, SQL). For
                `False` firstly is policy optimized then critic.
                defaults to False
            action_bounds (Union[list, np.ndarray, None], optional):
                action bounds. Applied for every generated action as
                clip, defaults to None
            sampling_time (float, optional): time interval between two
                consecutive actions, defaults to 0.1
        """
        Scenario.__init__(
            self,
            simulator=simulator,
            sampling_time=sampling_time,
            policy=policy,
            running_objective=running_objective,
            constraint_parser=constraint_parser,
            observer=observer,
            N_episodes=N_episodes,
            N_iterations=N_iterations,
            value_threshold=value_threshold,
            discount_factor=discount_factor,
        )
        is_parallel = self._metadata["argv"].parallel

        self.critic_optimization_event = critic_optimization_event
        self.policy_optimization_event = policy_optimization_event
        self.data_buffer = DataBuffer()
        self.critic = critic
        self.is_first_compute_action_call = True
        self.is_critic_first = is_critic_first
        self.stopping_criterion = stopping_criterion
        self.is_parallel = is_parallel
        self.seed_increment = 0
        self.annealing_exploration_factor = 1

    def _reset_whatever(self, suffix):
        def wrapped(*args, **kwargs):
            res = object.__getattribute__(self, f"reset_{suffix}")(*args, **kwargs)
            if self.sim_status != "simulation_ended":
                self.optimize(event=getattr(Event, f"reset_{suffix}"))
            if suffix == "iteration":
                self.data_buffer.nullify_buffer()
            return res

        return wrapped

    def run_episode(self, episode_counter, iteration_counter):
        super().run_episode(
            episode_counter=episode_counter, iteration_counter=iteration_counter
        )
        return self.data_buffer

    def run_ith_scenario(
        self, episode_id: int, iteration_id: int, scenarios: List[Self], queue
    ):
        seed = torch.initial_seed() + episode_id
        torch.manual_seed(seed)
        # np.random.seed(seed)
        random.seed(seed)
        self.seed_increment += 1

        queue.put(
            (
                episode_id,
                scenarios[episode_id - 1].run_episode(episode_id, iteration_id),
            )
        )

    def instantiate_rl_scenarios(self):
        simulators = [deepcopy(self.simulator) for _ in range(self.N_episodes)]
        # for simulator in simulators:
        #     simulator.state_init = simulator.state + np.random.normal(
        #         0, 0.5, simulator.state.shape
        #     )
        #     simulator.state = simulator.state_init
        #     simulator.system.state = simulator.state_init
        #     simulator.observation = simulator.get_observation(
        #         time=simulator.time,
        #         state=simulator.state_init,
        #         inputs=simulator.action_init,
        #     )
        # simulator.system.state_init = simulator.state_init
        scenarios = [
            RLScenario(
                policy=self.policy,
                critic=self.critic,
                running_objective=self.running_objective,
                simulator=simulators[i],
                policy_optimization_event=self.policy_optimization_event,
                critic_optimization_event=self.critic_optimization_event,
                discount_factor=self.discount_factor,
                is_critic_first=self.is_critic_first,
                sampling_time=self.sampling_time,
                constraint_parser=self.constraint_parser,
                observer=self.observer,
                N_episodes=self.N_episodes,  # for correct logging
                N_iterations=self.N_iterations,  # for correct logging
                value_threshold=self.value_threshold,
                stopping_criterion=self.stopping_criterion,
            )
            for i in range(self.N_episodes)
        ]
        return scenarios

    @apply_callbacks()
    def dump_data_buffer(self, episode_id: int, data_buffer: DataBuffer):
        return episode_id, data_buffer

    def run(self):
        if not self.is_parallel:
            return super().run()

        for self.iteration_counter in range(1, self.N_iterations + 1):
            # self.policy.model.stds *= self.annealing_exploration_factor
            one_episode_rl_scenarios = self.instantiate_rl_scenarios()
            result_queue = mp.Queue()
            args = [
                (i, self.iteration_counter, one_episode_rl_scenarios, result_queue)
                for i in range(1, self.N_episodes + 1)
            ]
            processes = [
                mp.Process(target=self.run_ith_scenario, args=arg) for arg in args
            ]
            for p in processes:
                p.start()

            enumerated_data_buffers = sorted(
                [result_queue.get() for _ in range(self.N_episodes)], key=lambda x: x[0]
            )

            for p in processes:
                p.join()

            for (episode_idx, data_buffer), scenario in zip(
                enumerated_data_buffers, one_episode_rl_scenarios
            ):
                self.dump_data_buffer(episode_idx, data_buffer)
                self.data_buffer.concat(data_buffer)
                data = data_buffer.to_pandas(["running_objective", "time"])
                scenario.value = calculate_value(
                    data["running_objective"],
                    data["time"],
                    self.discount_factor,
                    self.sampling_time,
                )
                scenario.reload_scenario()

            self.reset_iteration()
            if self.sim_status == "simulation_ended":
                break

    def reset_iteration(self):
        super().reset_iteration()
        if self.stopping_criterion is not None:
            if self.stopping_criterion(self.data_buffer):
                self.sim_status = "simulation_ended"

    def __getattribute__(self, name):
        if name == "issue_action":
            return self._issue_action
        elif name.startswith("reset_"):
            suffix = name.split("_")[-1]
            return self._reset_whatever(suffix)
        else:
            return object.__getattribute__(self, name)

    def on_observation_received(self, time, estimated_state, observation):
        self.critic.receive_estimated_state(estimated_state)
        self.policy.receive_estimated_state(estimated_state)
        self.policy.receive_observation(observation)

        received_data = super().on_observation_received(
            time, estimated_state, observation
        )
        self.data_buffer.push_to_end(**received_data)

        return received_data

    def on_action_issued(self, observation):
        received_data = super().on_action_issued(observation)
        self.data_buffer.push_to_end(**received_data)
        return received_data

    @apply_callbacks()
    def compute_action(
        self,
        estimated_state,
        observation,
        time=0,
    ):
        # Check data consistency
        assert np.allclose(
            estimated_state, self.data_buffer.get_latest("estimated_state")
        )
        assert np.allclose(observation, self.data_buffer.get_latest("observation"))
        assert np.allclose(time, self.data_buffer.get_latest("time"))

        self.optimize(Event.compute_action)
        return self.get_action_from_policy()

    def optimize(self, event):
        if self.is_first_compute_action_call and event == Event.compute_action:
            self.optimize_or_skip_on_event(self.policy, event)
            self.issue_action(self.data_buffer.get_latest("observation"))
            self.is_first_compute_action_call = False
        elif self.is_critic_first:
            self.optimize_or_skip_on_event(self.critic, event)
            self.optimize_or_skip_on_event(self.policy, event)
            if event == Event.compute_action:
                self.issue_action(self.data_buffer.get_latest("observation"))
        else:
            self.optimize_or_skip_on_event(self.policy, event)
            if event == Event.compute_action:
                self.issue_action(self.data_buffer.get_latest("observation"))
            self.optimize_or_skip_on_event(self.critic, event)

    def optimize_or_skip_on_event(
        self,
        optimizable_object: Union[Critic, Policy],
        event: str,
    ):
        if (
            isinstance(optimizable_object, Critic)
            and event == self.critic_optimization_event
        ) or (
            isinstance(optimizable_object, Policy)
            and event == self.policy_optimization_event
        ):
            self.pre_optimize(
                optimizable_object=optimizable_object,
                event=event,
                time=self.data_buffer.get_latest("time"),
            )
            optimizable_object.optimize(self.data_buffer)

    @apply_callbacks()
    def pre_optimize(
        self,
        optimizable_object: Union[Critic, Policy],
        event: str,
        time: Optional[int] = None,
    ):
        if isinstance(optimizable_object, Critic):
            which = "Critic"
        elif isinstance(optimizable_object, Policy):
            which = "Policy"
        else:
            raise ValueError("optimizable object can be either Critic or Policy")

        return which, event, time, self.episode_counter, self.iteration_counter


class CALFScenario(RLScenario):
    def __init__(
        self,
        policy: Policy,
        critic: CriticCALF,
        safe_policy: Policy,
        simulator: Simulator,
        running_objective: RunningObjective,
        discount_factor: float,
        sampling_time: float,
        observer: Optional[Observer] = None,
        N_iterations: int = 5,
        critic_learning_norm_threshold: float = 0.1,
        store_weights_thr: float = 0.0,
        weighted_norm_coeffs: Optional[List[float]] = None,
        weights_disturbance_std_after_iteration: Optional[float] = None,
        is_mean_weighted: bool = False,
        is_safe_filter_on: bool = True,
    ):
        super().__init__(
            policy=policy,
            critic=critic,
            simulator=simulator,
            running_objective=running_objective,
            critic_optimization_event=Event.compute_action,
            policy_optimization_event=Event.compute_action,
            discount_factor=discount_factor,
            sampling_time=sampling_time,
            observer=observer,
            N_iterations=N_iterations,
        )
        self.is_mean_weighted = is_mean_weighted
        self.safe_policy = safe_policy
        self.critic_learning_norm_threshold = critic_learning_norm_threshold
        self.weighted_norm_coeffs = (
            np.array(weighted_norm_coeffs)
            if weighted_norm_coeffs is not None
            else np.ones(self.policy.system.dim_observation)
        )
        self.store_weights_thr = store_weights_thr
        if store_weights_thr > 0.0:
            self.store_weights_for_next_iteration_trigger = (
                lambda state, thr: np.linalg.norm(state * self.weighted_norm_coeffs)
                > thr
            )
        else:
            self.store_weights_for_next_iteration_trigger = lambda state, thr: True

        self.stored_weights = self.critic.weights
        self.weights_disturbance_std_after_iteration = (
            weights_disturbance_std_after_iteration
        )
        self.is_safe_filter_on = is_safe_filter_on

    def issue_action(self, observation, is_safe=False):
        if is_safe:
            self.policy.restore_weights()
            self.critic.restore_weights()
            safe_action = self.safe_policy.get_action(observation)
            self.policy.set_action(safe_action)
        else:
            self.policy.update_action(observation)

    def on_action_issued(self, observation):
        data = super().on_action_issued(observation)
        data |= {"observation_last_good": self.critic.observation_last_good}
        self.data_buffer.push_to_end(
            observation_last_good=self.critic.observation_last_good
        )

        return data

    def compute_action_sampled(self, time, estimated_state, observation):
        self.is_time_for_new_sample = self.clock.check_time(time)
        if self.is_time_for_new_sample:
            self.on_observation_received(time, estimated_state, observation)
            action = self.simulator.system.apply_action_bounds(
                self.compute_action(
                    time=time,
                    estimated_state=estimated_state,
                    observation=observation,
                    is_update_critic=np.linalg.norm(
                        observation * self.weighted_norm_coeffs
                    )
                    > self.critic_learning_norm_threshold,
                )
            )
            self.post_compute_action(observation, estimated_state)
            self.step_counter += 1
            self.action_old = action
        else:
            action = self.action_old
        return action

    @apply_callbacks()
    def compute_action(
        self, estimated_state, observation, time=0, is_update_critic=True
    ):
        assert np.allclose(
            estimated_state, self.data_buffer.get_latest("estimated_state")
        )
        assert np.allclose(observation, self.data_buffer.get_latest("observation"))
        assert np.allclose(time, self.data_buffer.get_latest("time"))

        if self.is_first_compute_action_call:
            self.critic.observation_last_good = observation
            self.weights_for_next_iteration = self.critic.weights
            self.issue_action(observation, is_safe=True)
            self.is_first_compute_action_call = False
        else:
            self.pre_optimize(self.critic, Event.compute_action, time)
            if is_update_critic:
                critic_weights = self.critic.optimize(
                    self.data_buffer, is_update_and_cache_weights=False
                )
                critic_weights_accepted = (
                    self.critic.opt_status == OptStatus.success
                ) or not self.is_safe_filter_on

            else:
                critic_weights_accepted = False
                critic_weights = self.critic.weights
                self.critic.opt_status = OptStatus.failed

            if critic_weights_accepted:
                self.critic.update_weights(critic_weights)
                if self.store_weights_for_next_iteration_trigger(
                    observation, self.store_weights_thr
                ):
                    # print("STORED")
                    self.stored_weights = critic_weights
                self.pre_optimize(self.policy, Event.compute_action, time)
                self.policy.optimize(self.data_buffer)
                policy_weights_accepted = (
                    self.policy.opt_status == OptStatus.success
                ) or not self.is_safe_filter_on

                if policy_weights_accepted:
                    self.critic.observation_last_good = observation
                    self.issue_action(observation, is_safe=False)
                    self.critic.cache_weights(critic_weights)
                else:
                    self.issue_action(observation, is_safe=True)
            else:
                self.issue_action(observation, is_safe=True)

            step_id = self.data_buffer.get_latest("step_id")
            self.weights_for_next_iteration = (
                self.weights_for_next_iteration * (step_id - 1)
                + self.critic.weights * np.exp(-step_id)
            ) / step_id

        return self.get_action_from_policy()

    def reset_iteration(self):
        super().reset_iteration()
        if self.stored_weights is not None and self.store_weights_thr > 0.0:
            self.critic.model.update_and_cache_weights(self.stored_weights)
        self.critic.observation_last_good = self.simulator.observation_init
        # if self.weights_disturbance_std_after_iteration is not None:
        #     (
        #         self.critic.model.update_and_cache_weights(
        #             self.critic.weights
        #             + abs(np.random.randn(self.critic.model.weights.shape[0]))
        #             * self.weights_disturbance_std_after_iteration
        #         )
        #     )
        if self.is_mean_weighted:
            self.critic.model.update_and_cache_weights(self.weights_for_next_iteration)


class CALF(CALFScenario):
    """Scenario for CALF algorithm."""

    def __init__(
        self,
        simulator: Simulator,
        running_objective: RunningObjective,
        safe_policy: Policy,
        critic_td_n: int = 2,
        observer: Optional[Observer] = None,
        prediction_horizon: int = 1,
        policy_model: Optional[Model] = None,
        critic_model: Optional[Model] = None,
        predictor: Optional[Predictor] = None,
        discount_factor=1.0,
        sampling_time: float = 0.1,
        critic_lb_parameter: float = 0.0,
        critic_ub_parameter: float = 1.0,
        critic_safe_decay_param: float = 0.001,
        critic_is_dynamic_decay_rate: bool = False,
        critic_batch_size: int = 10,
        critic_regularization_param: float = 0,
        critic_learning_norm_threshold: float = 0.0,
        N_iterations=5,
        store_weights_thr=0,
        weighted_norm_coeffs: Optional[List[float]] = None,
        weights_disturbance_std_after_iteration: Optional[float] = None,
        is_mean_weighted: bool = False,
        is_safe_filter_on: bool = True,
    ):
        """Instantiate CALF class.

        Args:
            simulator (Simulator): The simulator object.
            running_objective (RunningObjective): The running objective.
            safe_policy (Policy): The safe policy.
            critic_td_n (int, optional): The TD-N parameter for the
                critic. Defaults to 2.
            observer (Optional[Observer], optional): The observer
                object. Defaults to None.
            prediction_horizon (int, optional): The prediction horizon.
                Defaults to 1.
            policy_model (Optional[Model], optional): The policy model.
                Defaults to None.
            critic_model (Optional[Model], optional): The critic model.
                Defaults to None.
            predictor (Optional[Predictor], optional): The predictor
                object. Defaults to None.
            discount_factor (float, optional): The discount factor.
                Defaults to 1.0.
            sampling_time (float, optional): The sampling time. Defaults
                to 0.1.
            critic_lb_parameter (float, optional): The lower bound
                parameter for the critic. Defaults to 0.0.
            critic_ub_parameter (float, optional): The upper bound
                parameter for the critic. Defaults to 1.0.
            critic_safe_decay_param (float, optional): The safe decay
                parameter for the critic. Defaults to 0.001.
            critic_is_dynamic_decay_rate (bool, optional): Whether the
                critic has a dynamic decay rate. Defaults to False.
            critic_batch_size (int, optional): The batch size for the
                critic optimizer. Defaults to 10.

        Returns:
            None: None
        """
        system = simulator.system
        critic = CriticCALF(
            system=system,
            model=critic_model,
            td_n=critic_td_n,
            is_same_critic=False,
            is_value_function=True,
            discount_factor=1.0,
            sampling_time=sampling_time,
            safe_decay_param=critic_safe_decay_param,
            is_dynamic_decay_rate=critic_is_dynamic_decay_rate,
            safe_policy=safe_policy,
            lb_parameter=critic_lb_parameter,
            ub_parameter=critic_ub_parameter,
            optimizer_config=CasadiOptimizerConfig(critic_batch_size),
            regularization_param=critic_regularization_param,
        )
        policy = RLPolicy(
            action_bounds=system.action_bounds,
            model=(
                ModelWeightContainer(
                    weights_init=np.zeros(
                        (prediction_horizon, system.dim_inputs),
                        dtype=np.float64,
                    ),
                    dim_output=system.dim_inputs,
                )
                if policy_model is None
                else policy_model
            ),
            system=system,
            running_objective=running_objective,
            prediction_horizon=prediction_horizon,
            algorithm="rpv",
            critic=critic,
            predictor=(
                predictor
                if predictor is not None
                else EulerPredictor(system=system, pred_step_size=sampling_time)
            ),
            discount_factor=discount_factor,
            optimizer_config=CasadiOptimizerConfig(),
        )

        super().__init__(
            policy=policy,
            critic=critic,
            safe_policy=safe_policy,
            simulator=simulator,
            running_objective=running_objective,
            discount_factor=discount_factor,
            sampling_time=sampling_time,
            observer=observer,
            N_iterations=N_iterations,
            critic_learning_norm_threshold=critic_learning_norm_threshold,
            store_weights_thr=store_weights_thr,
            weighted_norm_coeffs=weighted_norm_coeffs,
            weights_disturbance_std_after_iteration=weights_disturbance_std_after_iteration,
            is_mean_weighted=is_mean_weighted,
            is_safe_filter_on=is_safe_filter_on,
        )


class CALFTorch(CALFScenario):
    def __init__(
        self,
        simulator: Simulator,
        running_objective: RunningObjective,
        safe_policy: Policy,
        policy_n_epochs: int,
        critic_n_epochs: int,
        critic_n_epochs_per_constraint: int,
        policy_opt_method_kwargs: Dict[str, Any],
        critic_opt_method_kwargs: Dict[str, Any],
        policy_opt_method: Optional[Type[torch.optim.Optimizer]] = torch.optim.Adam,
        critic_opt_method: Optional[Type[torch.optim.Optimizer]] = torch.optim.Adam,
        critic_td_n: int = 2,
        observer: Optional[Observer] = None,
        prediction_horizon: int = 1,
        policy_model: Optional[Model] = None,
        critic_model: Optional[Model] = None,
        predictor: Optional[Predictor] = None,
        discount_factor=1.0,
        sampling_time: float = 0.1,
        critic_lb_parameter: float = 0.0,
        critic_ub_parameter: float = 1.0,
        critic_safe_decay_param: float = 0.001,
        critic_is_dynamic_decay_rate: bool = False,
        critic_batch_size: int = 10,
        N_iterations=5,
    ):
        system = simulator.system
        critic = CriticCALF(
            system=system,
            model=critic_model,
            td_n=critic_td_n,
            is_same_critic=False,
            is_value_function=True,
            discount_factor=1.0,
            sampling_time=sampling_time,
            safe_decay_param=critic_safe_decay_param,
            is_dynamic_decay_rate=critic_is_dynamic_decay_rate,
            safe_policy=safe_policy,
            lb_parameter=critic_lb_parameter,
            ub_parameter=critic_ub_parameter,
            optimizer_config=TorchOptimizerConfig(
                n_epochs=critic_n_epochs,
                opt_method=critic_opt_method,
                opt_method_kwargs=critic_opt_method_kwargs,
                data_buffer_iter_bathes_kwargs=dict(
                    batch_sampler=RollingBatchSampler,
                    dtype=torch.FloatTensor,
                    mode="backward",
                    batch_size=critic_batch_size,
                ),
                n_epochs_per_constraint=critic_n_epochs_per_constraint,
            ),
        )
        policy = RLPolicy(
            action_bounds=system.action_bounds,
            model=(
                ModelWeightContainerTorch(
                    dim_weights=(prediction_horizon, system.dim_inputs),
                    output_bounds=system.action_bounds,
                    output_bounding_type="clip",
                )
                if policy_model is None
                else policy_model
            ),
            system=system,
            running_objective=running_objective,
            prediction_horizon=prediction_horizon,
            algorithm="rpv",
            critic=critic,
            predictor=(
                predictor
                if predictor is not None
                else EulerPredictor(system=system, pred_step_size=sampling_time)
            ),
            discount_factor=discount_factor,
            optimizer_config=TorchOptimizerConfig(
                n_epochs=policy_n_epochs,
                opt_method=policy_opt_method,
                opt_method_kwargs=policy_opt_method_kwargs,
                data_buffer_iter_bathes_kwargs=dict(
                    batch_sampler=RollingBatchSampler,
                    dtype=torch.FloatTensor,
                    mode="backward",
                    batch_size=1,
                    n_batches=1,
                ),
            ),
        )

        super().__init__(
            policy=policy,
            critic=critic,
            safe_policy=safe_policy,
            simulator=simulator,
            running_objective=running_objective,
            discount_factor=discount_factor,
            sampling_time=sampling_time,
            observer=observer,
            N_iterations=N_iterations,
        )


class MPC(RLScenario):
    """Leverages the Model Predictive Control Scenario.

    The MPCScenario leverages the Model Predictive Control (MPC) approach within the reinforcement learning scenario,
    utilizing a prediction model to plan and apply sequences of actions that optimize the desired objectives over a time horizon.
    """

    def __init__(
        self,
        running_objective: RunningObjective,
        simulator: Simulator,
        prediction_horizon: int,
        predictor: Optional[Predictor] = None,
        sampling_time: float = 0.1,
        observer: Optional[Observer] = None,
        constraint_parser: Optional[ConstraintParser] = None,
        discount_factor: float = 1.0,
    ):
        """Initialize the MPC agent, setting up the required structures for MPC.

        Args:
            running_objective (RunningObjective): The objective function
                to assess the costs over the prediction horizon.
            simulator (Simulator): The environment simulation for
                applying and testing the agent.
            prediction_horizon (int): The number of steps into the
                future over which predictions are made.
            predictor (Optional[Predictor]): The prediction model used
                for forecasting future states.
            sampling_time (float): The time step interval for scenario.
            observer (Observer | None): The component for estimating the
                system's current state. Defaults to None.
            constraint_parser (Optional[ConstraintParser]): The
                mechanism for enforcing operational constraints.
                Defaults to None.
            discount_factor (float): The factor for discounting the
                value of future costs. Defaults to 1.0.
        """
        system = simulator.system
        super().__init__(
            N_episodes=1,
            N_iterations=1,
            simulator=simulator,
            policy_optimization_event=Event.compute_action,
            critic=CriticTrivial(),
            running_objective=running_objective,
            observer=observer,
            sampling_time=sampling_time,
            policy=RLPolicy(
                action_bounds=system.action_bounds,
                model=ModelWeightContainer(
                    weights_init=np.zeros(
                        (prediction_horizon + 1, system.dim_inputs), dtype=np.float64
                    ),
                    dim_output=system.dim_inputs,
                ),
                constraint_parser=constraint_parser,
                system=system,
                running_objective=running_objective,
                prediction_horizon=prediction_horizon,
                algorithm="mpc",
                critic=CriticTrivial(),
                predictor=(
                    predictor
                    if predictor is not None
                    else EulerPredictor(system=system, pred_step_size=sampling_time)
                ),
                discount_factor=discount_factor,
                optimizer_config=CasadiOptimizerConfig(),
            ),
        )


def get_predictive_kwargs(
    running_objective: RunningObjective,
    simulator: Simulator,
    prediction_horizon: int,
    predictor: Predictor,
    sampling_time: float,
    observer: Optional[Observer],
    constraint_parser: Optional[ConstraintParser],
    N_iterations: int,
    discount_factor: float,
    algorithm_name: str,
    stopping_criterion: Optional[Callable[[DataBuffer], bool]] = None,
    policy_model: Optional[Model] = None,
    critic_td_n: Optional[int] = None,
    critic_is_on_policy: Optional[bool] = None,
    critic_is_value_function: Optional[bool] = None,
    critic_regularization_param: Optional[float] = None,
    critic_is_same_critic: Optional[bool] = None,
    critic_model: Optional[Model] = None,
    critic_n_samples: Optional[int] = None,
    epsilon_random_parameter: Optional[float] = None,
):
    system = simulator.system
    critic = Critic(
        system=system,
        model=(
            critic_model
            if critic_model is not None
            else ModelQuadLin("diagonal", is_with_linear_terms=False)
        ),
        td_n=critic_td_n,
        is_on_policy=critic_is_on_policy,
        is_value_function=critic_is_value_function,
        sampling_time=sampling_time,
        discount_factor=discount_factor,
        regularization_param=critic_regularization_param,
        action_bounds=system.action_bounds,
        is_same_critic=critic_is_same_critic,
        optimizer_config=CasadiOptimizerConfig(batch_size=critic_n_samples),
    )

    policy = RLPolicy(
        action_bounds=system.action_bounds,
        model=(
            ModelWeightContainer(
                weights_init=np.zeros(
                    (
                        prediction_horizon + (1 if algorithm_name != "rpv" else 0),
                        system.dim_inputs,
                    ),
                    dtype=np.float64,
                ),
                dim_output=system.dim_inputs,
            )
            if policy_model is None
            else policy_model
        ),
        constraint_parser=constraint_parser,
        system=system,
        running_objective=running_objective,
        prediction_horizon=prediction_horizon,
        algorithm=algorithm_name,
        critic=critic,
        predictor=(
            predictor
            if predictor is not None
            else EulerPredictor(system=system, pred_step_size=sampling_time)
        ),
        discount_factor=discount_factor,
        optimizer_config=CasadiOptimizerConfig(batch_size=1),
        epsilon_random_parameter=epsilon_random_parameter,
    )

    return dict(
        stopping_criterion=stopping_criterion,
        N_episodes=1,
        N_iterations=N_iterations,
        simulator=simulator,
        policy_optimization_event=Event.compute_action,
        critic_optimization_event=Event.compute_action,
        critic=critic,
        running_objective=running_objective,
        observer=observer,
        sampling_time=sampling_time,
        policy=policy,
        is_critic_first=True,
    )


class SQL(RLScenario):
    """Implements Stacked Q-Learning algorithm."""

    def __init__(
        self,
        running_objective: RunningObjective,
        simulator: Simulator,
        prediction_horizon: int,
        critic_td_n: int,
        critic_batch_size: int,
        predictor: Optional[Predictor] = None,
        sampling_time: float = 0.1,
        observer: Optional[Observer] = None,
        constraint_parser: Optional[ConstraintParser] = None,
        N_iterations: int = 1,
        discount_factor: float = 1.0,
        policy_model: Optional[Model] = None,
        critic_model: Optional[Model] = None,
        critic_regularization_param: float = 0.0,
        epsilon_random_parameter: Optional[float] = None,
        stopping_criterion: Optional[Callable[[DataBuffer], bool]] = None,
    ):
        """Instantiate SQL.

        Args:
            running_objective (RunningObjective): The objective function
                to assess the costs over the prediction horizon.
            simulator (Simulator): The environment simulation for
                applying and testing the agent.
            prediction_horizon (int): The number of steps into the
                future over which predictions are made.
            critic_td_n (int): The n-step temporal-difference parameter
                used for critic updates.
            critic_batch_size (int): The batch size for the critic.
            predictor (Optional[Predictor]): The prediction model used
                for forecasting future states. Defaults to None.
            sampling_time (float): The time step interval for scenario.
                Defaults to 0.1.
            observer (Optional[Observer]): The observer object used for
                the algorithm. Defaults to None.
            constraint_parser (Optional[ConstraintParser]): The
                component for estimating the system's current state.
                Defaults to None.
            N_iterations (int): The number of iterations for the
                algorithm. Defaults to 1.
            discount_factor (float): The factor for discounting the
                value of future costs. Defaults to 1.0.
            policy_model (Optional[Model]): The model parameterizing the
                policy. Defaults to None. If `None` then
                `ModelWeightContainer` is used.
            critic_model (Optional[Model]): The model parameterizing the
                critic. Defaults to None. If `None` then diagonal
                quadratic form is used.
            critic_regularization_param (float): The regularization
                parameter for the critic. Defaults to 0.

        Returns:
            None: None
        """
        super().__init__(
            **get_predictive_kwargs(
                simulator=simulator,
                running_objective=running_objective,
                prediction_horizon=prediction_horizon,
                predictor=predictor,
                sampling_time=sampling_time,
                observer=observer,
                constraint_parser=constraint_parser,
                N_iterations=N_iterations,
                discount_factor=discount_factor,
                algorithm_name="sql",
                policy_model=policy_model,
                critic_model=critic_model,
                critic_regularization_param=critic_regularization_param,
                critic_td_n=critic_td_n,
                critic_is_on_policy=True,
                critic_is_value_function=False,
                critic_is_same_critic=False,
                critic_n_samples=critic_batch_size,
                epsilon_random_parameter=epsilon_random_parameter,
                stopping_criterion=stopping_criterion,
            )
        )


class RQL(RLScenario):
    """Implements Rollout Q-Learning algorithm."""

    def __init__(
        self,
        running_objective: RunningObjective,
        simulator: Simulator,
        prediction_horizon: int,
        critic_td_n: int,
        critic_batch_size: int,
        predictor: Optional[Predictor] = None,
        sampling_time: float = 0.1,
        observer: Optional[Observer] = None,
        constraint_parser: Optional[ConstraintParser] = None,
        N_iterations: int = 1,
        discount_factor: float = 1.0,
        policy_model: Optional[Model] = None,
        critic_model: Optional[Model] = None,
        critic_regularization_param: float = 0.0,
        epsilon_random_parameter: Optional[float] = None,
        stopping_criterion: Optional[Callable[[DataBuffer], bool]] = None,
    ):
        """Instantiate RQLScenario.

        Args:
            running_objective (RunningObjective): The objective function
                to assess the costs over the prediction horizon.
            simulator (Simulator): The environment simulation for
                applying and testing the agent.
            prediction_horizon (int): The number of steps into the
                future over which predictions are made.
            critic_td_n (int): The n-step temporal-difference parameter
                used for critic updates.
            critic_batch_size (int): The batch size for the critic.
            predictor (Optional[Predictor]): The prediction model used
                for forecasting future states. Defaults to None.
            sampling_time (float): The time step interval for scenario.
                Defaults to 0.1.
            observer (Optional[Observer]): The observer object used for
                the algorithm. Defaults to None.
            constraint_parser (Optional[ConstraintParser]): The
                component for estimating the system's current state.
                Defaults to None.
            N_iterations (int): The number of iterations for the
                algorithm. Defaults to 1.
            discount_factor (float): The factor for discounting the
                value of future costs. Defaults to 1.0.
            policy_model (Optional[Model]): The model parameterizing the
                policy. Defaults to None. If `None` then
                `ModelWeightContainer` is used.
            critic_model (Optional[Model]): The model parameterizing the
                critic. Defaults to None. If `None` then diagonal
                quadratic form is used.
            critic_regularization_param (float): The regularization
                parameter for the critic. Defaults to 0.

        Returns:
            None: None
        """
        super().__init__(
            **get_predictive_kwargs(
                stopping_criterion=stopping_criterion,
                simulator=simulator,
                running_objective=running_objective,
                prediction_horizon=prediction_horizon,
                predictor=predictor,
                sampling_time=sampling_time,
                observer=observer,
                constraint_parser=constraint_parser,
                N_iterations=N_iterations,
                discount_factor=discount_factor,
                algorithm_name="rql",
                policy_model=policy_model,
                critic_model=critic_model,
                critic_regularization_param=critic_regularization_param,
                critic_td_n=critic_td_n,
                critic_is_on_policy=True,
                critic_is_value_function=False,
                critic_is_same_critic=False,
                critic_n_samples=critic_batch_size,
                epsilon_random_parameter=epsilon_random_parameter,
            )
        )


class RPV(RLScenario):
    """Implements Reward + Value algorithm."""

    def __init__(
        self,
        running_objective: RunningObjective,
        simulator: Simulator,
        prediction_horizon: int,
        critic_td_n: int,
        critic_batch_size: int,
        predictor: Optional[Predictor] = None,
        sampling_time: float = 0.1,
        observer: Optional[Observer] = None,
        constraint_parser: Optional[ConstraintParser] = None,
        N_iterations: int = 1,
        discount_factor: float = 1.0,
        policy_model: Optional[Model] = None,
        critic_model: Optional[Model] = None,
        critic_regularization_param: float = 0.0,
        stopping_criterion: Optional[Callable[[DataBuffer], bool]] = None,
    ):
        """Instantiate RPV.

        Args:
            running_objective (RunningObjective): The objective function
                to assess the costs over the prediction horizon.
            simulator (Simulator): The environment simulation for
                applying and testing the agent.
            prediction_horizon (int): The number of steps into the
                future over which predictions are made.
            critic_td_n (int): The n-step temporal-difference parameter
                used for critic updates.
            critic_batch_size (int): The batch size for the critic.
            predictor (Optional[Predictor]): The prediction model used
                for forecasting future states. Defaults to None.
            sampling_time (float): The time step interval for scenario.
                Defaults to 0.1.
            observer (Optional[Observer]): The observer object used for
                the algorithm. Defaults to None.
            constraint_parser (Optional[ConstraintParser]): The
                component for estimating the system's current state.
                Defaults to None.
            N_iterations (int): The number of iterations for the
                algorithm. Defaults to 1.
            discount_factor (float): The factor for discounting the
                value of future costs. Defaults to 1.0.
            policy_model (Optional[Model]): The model parameterizing the
                policy. Defaults to None. If `None` then
                `ModelWeightContainer` is used.
            critic_model (Optional[Model]): The model parameterizing the
                critic. Defaults to None. If `None` then diagonal
                quadratic form is used.
            critic_regularization_param (float): The regularization
                parameter for the critic. Defaults to 0.

        Returns:
            None: None
        """
        super().__init__(
            **get_predictive_kwargs(
                stopping_criterion=stopping_criterion,
                simulator=simulator,
                running_objective=running_objective,
                prediction_horizon=prediction_horizon,
                predictor=predictor,
                sampling_time=sampling_time,
                observer=observer,
                constraint_parser=constraint_parser,
                N_iterations=N_iterations,
                discount_factor=discount_factor,
                algorithm_name="rpv",
                policy_model=policy_model,
                critic_model=critic_model,
                critic_regularization_param=critic_regularization_param,
                critic_td_n=critic_td_n,
                critic_is_on_policy=True,
                critic_is_value_function=True,
                critic_is_same_critic=False,
                critic_n_samples=critic_batch_size,
            )
        )


class MPCTorch(RLScenario):
    """MPCTorchScenario encapsulates the model predictive control (MPC) approach using PyTorch optimization for reinforcement learning.

    It integrates a policy that employs model-based predictions over a specified horizon to optimize actions, taking into account the dynamic nature of the environment. The scenario coordinates the interaction between the policy, model, critic, and environment.
    """

    def __init__(
        self,
        running_objective: RunningObjective,
        simulator: Simulator,
        prediction_horizon: int,
        sampling_time: float,
        n_epochs: int,
        opt_method_kwargs: Dict[str, Any],
        opt_method: Type[torch.optim.Optimizer] = torch.optim.Adam,
        predictor: Optional[Predictor] = None,
        model: Optional[ModelNN] = None,
        observer: Optional[Observer] = None,
        constraint_parser: Optional[ConstraintParser] = None,
        discount_factor: float = 1.0,
    ):
        """Initialize the object with the given parameters.

        Args:
            running_objective (RunningObjective): The objective function
                to assess the costs over the prediction horizon.
            simulator (Simulator): The environment simulation for
                applying and testing the agent.
            prediction_horizon (int): The number of steps into the
                future over which predictions are made.
            sampling_time (float): The time step interval for scenario.
            n_epochs (int): The number of training epochs.
            opt_method_kwargs (Dict[str, Any]): Additional keyword
                arguments for the optimization method.
            opt_method (Type[torch.optim.Optimizer]): The optimization
                method to use. Defaults to torch.optim.Adam.
            predictor (Optional[Predictor]): The predictor object to
                use. Defaults to None.
            model (Optional[ModelNN]): The neural network model
                parameterizing the policy. Defaults to None. If `None`
                then `ModelWeightContainerTorch` is used, but you can
                implement any neural network you want.
            observer (Optional[Observer]): Object responsible for state
                estimation from observations. Defaults to None.
            constraint_parser (Optional[ConstraintParser]): The
                mechanism for enforcing operational constraints.
                Defaults to None.
            discount_factor (float): The discount factor for future
                costs. Defaults to 1.0.

        Returns:
            None
        """
        system = simulator.system
        super().__init__(
            simulator=simulator,
            N_episodes=1,
            N_iterations=1,
            policy_optimization_event=Event.compute_action,
            critic=CriticTrivial(),
            running_objective=running_objective,
            observer=observer,
            sampling_time=sampling_time,
            policy=RLPolicy(
                action_bounds=system.action_bounds,
                model=(
                    ModelWeightContainerTorch(
                        dim_weights=(prediction_horizon + 1, system.dim_inputs),
                        output_bounds=system.action_bounds,
                    )
                    if model is None
                    else model
                ),
                constraint_parser=constraint_parser,
                system=system,
                running_objective=running_objective,
                prediction_horizon=prediction_horizon,
                algorithm="mpc",
                critic=CriticTrivial(),
                predictor=(
                    predictor
                    if predictor is not None
                    else EulerPredictor(system=system, pred_step_size=sampling_time)
                ),
                discount_factor=discount_factor,
                optimizer_config=TorchOptimizerConfig(
                    n_epochs=n_epochs,
                    opt_method=opt_method,
                    opt_method_kwargs=opt_method_kwargs,
                    data_buffer_iter_bathes_kwargs=dict(
                        batch_sampler=RollingBatchSampler,
                        dtype=torch.FloatTensor,
                        mode="backward",
                        batch_size=1,
                    ),
                ),
            ),
        )


def get_predictive_torch_kwargs(
    running_objective: RunningObjective,
    simulator: Simulator,
    prediction_horizon: int,
    predictor: Predictor,
    sampling_time: float,
    observer: Optional[Observer],
    constraint_parser: Optional[ConstraintParser],
    N_iterations: int,
    discount_factor: float,
    algorithm_name: str,
    policy_model: Optional[ModelNN] = None,
    policy_n_epochs: Optional[int] = None,
    policy_opt_method: Optional[Type[torch.optim.Optimizer]] = torch.optim.Adam,
    policy_opt_method_kwargs: Optional[Dict[str, Any]] = None,
    critic_td_n: Optional[int] = None,
    critic_is_on_policy: Optional[bool] = None,
    critic_is_value_function: Optional[bool] = None,
    critic_is_same_critic: Optional[bool] = None,
    critic_model: Optional[ModelNN] = None,
    critic_n_epochs: Optional[int] = None,
    critic_opt_method: Optional[Type[torch.optim.Optimizer]] = torch.optim.Adam,
    critic_opt_method_kwargs: Optional[Dict[str, Any]] = None,
    critic_batch_size: Optional[int] = None,
    epsilon_random_parameter: Optional[float] = None,
    size_mesh: Optional[int] = None,
    stopping_criterion: Optional[Callable[[DataBuffer], bool]] = None,
):
    system = simulator.system
    critic = Critic(
        system=system,
        model=critic_model,
        td_n=critic_td_n,
        is_on_policy=critic_is_on_policy,
        is_value_function=critic_is_value_function,
        sampling_time=sampling_time,
        discount_factor=discount_factor,
        action_bounds=system.action_bounds,
        is_same_critic=critic_is_same_critic,
        optimizer_config=TorchOptimizerConfig(
            n_epochs=critic_n_epochs,
            opt_method=critic_opt_method,
            opt_method_kwargs=critic_opt_method_kwargs,
            data_buffer_iter_bathes_kwargs=dict(
                batch_sampler=RollingBatchSampler,
                dtype=torch.FloatTensor,
                mode="backward",
                batch_size=critic_batch_size,
            ),
        ),
        size_mesh=size_mesh,
    )

    if algorithm_name == "greedy":
        predictor_argument = None
    elif predictor is None:
        predictor_argument = EulerPredictor(system=system, pred_step_size=sampling_time)
    else:
        predictor_argument = predictor

    policy = RLPolicy(
        action_bounds=system.action_bounds,
        model=(
            ModelWeightContainerTorch(
                dim_weights=(
                    prediction_horizon + (1 if algorithm_name != "rpv" else 0),
                    system.dim_inputs,
                ),
                output_bounds=system.action_bounds,
                output_bounding_type="clip",
            )
            if policy_model is None
            else policy_model
        ),
        constraint_parser=constraint_parser,
        system=system,
        running_objective=running_objective,
        prediction_horizon=prediction_horizon,
        algorithm=algorithm_name,
        critic=critic,
        predictor=predictor_argument,
        discount_factor=discount_factor,
        optimizer_config=TorchOptimizerConfig(
            n_epochs=policy_n_epochs,
            opt_method=policy_opt_method,
            opt_method_kwargs=policy_opt_method_kwargs,
            data_buffer_iter_bathes_kwargs=dict(
                batch_sampler=RollingBatchSampler,
                dtype=torch.FloatTensor,
                mode="backward",
                batch_size=1,
                n_batches=1,
            ),
        ),
        epsilon_random_parameter=epsilon_random_parameter,
    )

    return dict(
        N_episodes=1,
        N_iterations=N_iterations,
        simulator=simulator,
        policy_optimization_event=Event.compute_action,
        critic_optimization_event=Event.compute_action,
        critic=critic,
        running_objective=running_objective,
        observer=observer,
        sampling_time=sampling_time,
        policy=policy,
        is_critic_first=True,
        stopping_criterion=stopping_criterion,
    )


class SQLTorch(RLScenario):
    def __init__(
        self,
        running_objective: RunningObjective,
        simulator: Simulator,
        prediction_horizon: int,
        critic_td_n: int,
        critic_batch_size: int,
        policy_n_epochs: int,
        critic_n_epochs: int,
        policy_opt_method_kwargs: Optional[Dict[str, Any]],
        critic_opt_method_kwargs: Optional[Dict[str, Any]],
        predictor: Optional[Predictor] = None,
        sampling_time: float = 0.1,
        observer: Optional[Observer] = None,
        constraint_parser: Optional[ConstraintParser] = None,
        N_iterations: int = 1,
        discount_factor: float = 1.0,
        policy_model: Optional[ModelNN] = None,
        policy_opt_method: Optional[Type[torch.optim.Optimizer]] = torch.optim.Adam,
        critic_opt_method: Optional[Type[torch.optim.Optimizer]] = torch.optim.Adam,
        critic_model: Optional[ModelNN] = None,
        stopping_criterion: Optional[Callable[[DataBuffer], bool]] = None,
    ):
        super().__init__(
            **get_predictive_torch_kwargs(
                running_objective=running_objective,
                simulator=simulator,
                prediction_horizon=prediction_horizon,
                predictor=predictor,
                sampling_time=sampling_time,
                observer=observer,
                constraint_parser=constraint_parser,
                N_iterations=N_iterations,
                discount_factor=discount_factor,
                algorithm_name="sql",
                policy_model=policy_model,
                policy_n_epochs=policy_n_epochs,
                policy_opt_method=policy_opt_method,
                policy_opt_method_kwargs=policy_opt_method_kwargs,
                critic_td_n=critic_td_n,
                critic_is_on_policy=True,
                critic_is_value_function=False,
                critic_is_same_critic=False,
                critic_model=critic_model,
                critic_n_epochs=critic_n_epochs,
                critic_opt_method=critic_opt_method,
                critic_opt_method_kwargs=critic_opt_method_kwargs,
                critic_batch_size=critic_batch_size,
                stopping_criterion=stopping_criterion,
            )
        )


class RQLTorch(RLScenario):
    def __init__(
        self,
        running_objective: RunningObjective,
        simulator: Simulator,
        prediction_horizon: int,
        critic_td_n: int,
        critic_batch_size: int,
        policy_n_epochs: int,
        critic_n_epochs: int,
        policy_opt_method_kwargs: Optional[Dict[str, Any]],
        critic_opt_method_kwargs: Optional[Dict[str, Any]],
        predictor: Optional[Predictor] = None,
        sampling_time: float = 0.1,
        observer: Optional[Observer] = None,
        constraint_parser: Optional[ConstraintParser] = None,
        N_iterations: int = 1,
        discount_factor: float = 1.0,
        policy_model: Optional[ModelNN] = None,
        policy_opt_method: Optional[Type[torch.optim.Optimizer]] = torch.optim.Adam,
        critic_opt_method: Optional[Type[torch.optim.Optimizer]] = torch.optim.Adam,
        critic_model: Optional[ModelNN] = None,
        epsilon_random_parameter: Optional[float] = None,
        stopping_criterion: Optional[Callable[[DataBuffer], bool]] = None,
    ):
        super().__init__(
            **get_predictive_torch_kwargs(
                running_objective=running_objective,
                simulator=simulator,
                prediction_horizon=prediction_horizon,
                predictor=predictor,
                sampling_time=sampling_time,
                observer=observer,
                constraint_parser=constraint_parser,
                N_iterations=N_iterations,
                discount_factor=discount_factor,
                algorithm_name="rql",
                policy_model=policy_model,
                policy_n_epochs=policy_n_epochs,
                policy_opt_method=policy_opt_method,
                policy_opt_method_kwargs=policy_opt_method_kwargs,
                critic_td_n=critic_td_n,
                critic_is_on_policy=True,
                critic_is_value_function=False,
                critic_is_same_critic=False,
                critic_model=critic_model,
                critic_n_epochs=critic_n_epochs,
                critic_opt_method=critic_opt_method,
                critic_opt_method_kwargs=critic_opt_method_kwargs,
                critic_batch_size=critic_batch_size,
                epsilon_random_parameter=epsilon_random_parameter,
                stopping_criterion=stopping_criterion,
            )
        )


class RPVTorch(RLScenario):
    def __init__(
        self,
        running_objective: RunningObjective,
        simulator: Simulator,
        prediction_horizon: int,
        critic_td_n: int,
        critic_batch_size: int,
        policy_n_epochs: int,
        critic_n_epochs: int,
        policy_opt_method_kwargs: Optional[Dict[str, Any]],
        critic_opt_method_kwargs: Optional[Dict[str, Any]],
        predictor: Optional[Predictor] = None,
        sampling_time: float = 0.1,
        observer: Optional[Observer] = None,
        constraint_parser: Optional[ConstraintParser] = None,
        N_iterations: int = 1,
        discount_factor: float = 1.0,
        policy_model: Optional[ModelNN] = None,
        policy_opt_method: Optional[Type[torch.optim.Optimizer]] = torch.optim.Adam,
        critic_opt_method: Optional[Type[torch.optim.Optimizer]] = torch.optim.Adam,
        critic_model: Optional[ModelNN] = None,
        epsilon_random_parameter: Optional[float] = None,
        stopping_criterion: Optional[Callable[[DataBuffer], bool]] = None,
    ):
        super().__init__(
            **get_predictive_torch_kwargs(
                running_objective=running_objective,
                simulator=simulator,
                prediction_horizon=prediction_horizon,
                predictor=predictor,
                sampling_time=sampling_time,
                observer=observer,
                constraint_parser=constraint_parser,
                N_iterations=N_iterations,
                discount_factor=discount_factor,
                algorithm_name="rpv",
                policy_model=policy_model,
                policy_n_epochs=policy_n_epochs,
                policy_opt_method=policy_opt_method,
                policy_opt_method_kwargs=policy_opt_method_kwargs,
                critic_td_n=critic_td_n,
                critic_is_on_policy=True,
                critic_is_value_function=True,
                critic_is_same_critic=False,
                critic_model=critic_model,
                critic_n_epochs=critic_n_epochs,
                critic_opt_method=critic_opt_method,
                critic_opt_method_kwargs=critic_opt_method_kwargs,
                critic_batch_size=critic_batch_size,
                epsilon_random_parameter=epsilon_random_parameter,
                stopping_criterion=stopping_criterion,
            )
        )


class SARSA(RLScenario):
    def __init__(
        self,
        running_objective: RunningObjective,
        simulator: Simulator,
        critic_td_n: int,
        critic_batch_size: int,
        policy_n_epochs: int,
        critic_n_epochs: int,
        critic_model: Optional[ModelNN],
        policy_opt_method_kwargs: Optional[Dict[str, Any]],
        critic_opt_method_kwargs: Optional[Dict[str, Any]],
        sampling_time: float = 0.1,
        observer: Optional[Observer] = None,
        constraint_parser: Optional[ConstraintParser] = None,
        N_iterations: int = 1,
        discount_factor: float = 1.0,
        policy_model: Optional[ModelNN] = None,
        policy_opt_method: Optional[Type[torch.optim.Optimizer]] = torch.optim.Adam,
        critic_opt_method: Optional[Type[torch.optim.Optimizer]] = torch.optim.Adam,
        epsilon_random_parameter: Optional[float] = None,
        stopping_criterion: Optional[Callable[[DataBuffer], bool]] = None,
    ):
        super().__init__(
            **get_predictive_torch_kwargs(
                running_objective=running_objective,
                simulator=simulator,
                prediction_horizon=0,
                predictor=None,
                sampling_time=sampling_time,
                observer=observer,
                constraint_parser=constraint_parser,
                N_iterations=N_iterations,
                discount_factor=discount_factor,
                algorithm_name="greedy",
                policy_model=policy_model,
                policy_n_epochs=policy_n_epochs,
                policy_opt_method=policy_opt_method,
                policy_opt_method_kwargs=policy_opt_method_kwargs,
                critic_td_n=critic_td_n,
                critic_is_on_policy=True,
                critic_is_value_function=False,
                critic_is_same_critic=False,
                critic_model=critic_model,
                critic_n_epochs=critic_n_epochs,
                critic_opt_method=critic_opt_method,
                critic_opt_method_kwargs=critic_opt_method_kwargs,
                critic_batch_size=critic_batch_size,
                epsilon_random_parameter=epsilon_random_parameter,
                stopping_criterion=stopping_criterion,
            )
        )


class DQN(RLScenario):
    def __init__(
        self,
        running_objective: RunningObjective,
        simulator: Simulator,
        critic_td_n: int,
        critic_batch_size: int,
        policy_n_epochs: int,
        critic_n_epochs: int,
        critic_model: Optional[ModelNN],
        policy_opt_method_kwargs: Optional[Dict[str, Any]],
        critic_opt_method_kwargs: Optional[Dict[str, Any]],
        sampling_time: float = 0.1,
        observer: Optional[Observer] = None,
        constraint_parser: Optional[ConstraintParser] = None,
        N_iterations: int = 1,
        discount_factor: float = 1.0,
        policy_model: Optional[ModelNN] = None,
        policy_opt_method: Optional[Type[torch.optim.Optimizer]] = torch.optim.Adam,
        critic_opt_method: Optional[Type[torch.optim.Optimizer]] = torch.optim.Adam,
        epsilon_random_parameter: Optional[float] = None,
        size_mesh: int = 100,
        stopping_criterion: Optional[Callable[[DataBuffer], bool]] = None,
    ):
        super().__init__(
            **get_predictive_torch_kwargs(
                running_objective=running_objective,
                simulator=simulator,
                prediction_horizon=0,
                predictor=None,
                sampling_time=sampling_time,
                observer=observer,
                constraint_parser=constraint_parser,
                N_iterations=N_iterations,
                discount_factor=discount_factor,
                algorithm_name="greedy",
                policy_model=policy_model,
                policy_n_epochs=policy_n_epochs,
                policy_opt_method=policy_opt_method,
                policy_opt_method_kwargs=policy_opt_method_kwargs,
                critic_td_n=critic_td_n,
                critic_is_on_policy=False,
                critic_is_value_function=False,
                critic_is_same_critic=False,
                critic_model=critic_model,
                critic_n_epochs=critic_n_epochs,
                critic_opt_method=critic_opt_method,
                critic_opt_method_kwargs=critic_opt_method_kwargs,
                critic_batch_size=critic_batch_size,
                epsilon_random_parameter=epsilon_random_parameter,
                size_mesh=size_mesh,
                stopping_criterion=stopping_criterion,
            )
        )


def get_policy_gradient_kwargs(
    sampling_time: float,
    running_objective: RunningObjective,
    simulator: Simulator,
    discount_factor: float,
    observer: Optional[Observer],
    N_episodes: int,
    N_iterations: int,
    value_threshold: float,
    policy_type: Type[Policy],
    policy_model: PerceptronWithTruncatedNormalNoise,
    policy_n_epochs: int,
    policy_opt_method_kwargs: Dict[str, Any],
    policy_opt_method: Type[torch.optim.Optimizer],
    is_reinstantiate_policy_optimizer: bool,
    critic_model: Optional[ModelPerceptron] = None,
    critic_opt_method: Optional[Type[torch.optim.Optimizer]] = None,
    critic_opt_method_kwargs: Optional[Dict[str, Any]] = None,
    critic_n_epochs: Optional[int] = None,
    critic_td_n: Optional[int] = None,
    critic_kwargs: Dict[str, Any] = None,
    critic_is_value_function: Optional[bool] = None,
    is_reinstantiate_critic_optimizer: Optional[bool] = None,
    policy_kwargs: Dict[str, Any] = None,
    scenario_kwargs: Dict[str, Any] = None,
    is_use_critic_as_policy_kwarg: bool = True,
    stopping_criterion: Optional[Callable[[DataBuffer], bool]] = None,
    device: str = "cpu",
):
    system = simulator.system
    if critic_model is not None:
        assert (
            critic_n_epochs is not None
            and critic_td_n is not None
            and critic_opt_method is not None
            and critic_opt_method_kwargs is not None
            and is_reinstantiate_critic_optimizer is not None
            and critic_is_value_function is not None
        ), "critic_n_epochs, critic_td_n, critic_opt_method, critic_opt_method_kwargs, is_reinstantiate_critic_optimizer, critic_is_value_function should be set"
        critic = Critic(
            system=system,
            discount_factor=discount_factor,
            sampling_time=sampling_time,
            model=critic_model,
            td_n=critic_td_n,
            device=device,
            is_value_function=critic_is_value_function,
            is_on_policy=True,
            is_same_critic=True,
            optimizer_config=TorchOptimizerConfig(
                critic_n_epochs,
                data_buffer_iter_bathes_kwargs={
                    "batch_sampler": RollingBatchSampler,
                    "dtype": torch.FloatTensor,
                    "mode": "full",
                    "n_batches": 1,
                    "device": device,
                },
                opt_method_kwargs=critic_opt_method_kwargs,
                opt_method=critic_opt_method,
                is_reinstantiate_optimizer=is_reinstantiate_critic_optimizer,
            ),
            is_full_iteration_epoch=True,
            **(critic_kwargs if critic_kwargs is not None else dict()),
        )
    else:
        critic = CriticTrivial()

    return dict(
        stopping_criterion=stopping_criterion,
        simulator=simulator,
        discount_factor=discount_factor,
        policy_optimization_event=Event.reset_iteration,
        critic_optimization_event=Event.reset_iteration,
        N_episodes=N_episodes,
        N_iterations=N_iterations,
        running_objective=running_objective,
        observer=observer,
        critic=critic,
        is_critic_first=True,
        value_threshold=value_threshold,
        sampling_time=sampling_time,
        policy=policy_type(
            model=policy_model,
            system=system,
            device=device,
            discount_factor=discount_factor,
            optimizer_config=TorchOptimizerConfig(
                n_epochs=policy_n_epochs,
                opt_method=policy_opt_method,
                opt_method_kwargs=policy_opt_method_kwargs,
                is_reinstantiate_optimizer=is_reinstantiate_policy_optimizer,
                data_buffer_iter_bathes_kwargs={
                    "batch_sampler": RollingBatchSampler,
                    "dtype": torch.FloatTensor,
                    "mode": "full",
                    "n_batches": 1,
                    "device": device,
                },
            ),
            **(dict(critic=critic) if is_use_critic_as_policy_kwarg else dict()),
            **(policy_kwargs if policy_kwargs is not None else dict()),
        ),
        **(scenario_kwargs if scenario_kwargs is not None else dict()),
    )


class PPO(RLScenario):
    """Scenario for Proximal Polizy Optimization.

    PPOScenario is a reinforcement learning scenario implementing the Proximal Policy Optimization (PPO) algorithm.
    This algorithm uses a policy gradient approach with an objective function designed to reduce the variance
    of policy updates, while ensuring that the new policy does not deviate significantly from the old one.
    """

    def __init__(
        self,
        policy_model: PerceptronWithTruncatedNormalNoise,
        critic_model: ModelPerceptron,
        sampling_time: float,
        running_objective: RunningObjective,
        simulator: Simulator,
        critic_n_epochs: int,
        policy_n_epochs: int,
        critic_opt_method_kwargs: Dict[str, Any],
        policy_opt_method_kwargs: Dict[str, Any],
        critic_opt_method: Type[torch.optim.Optimizer] = torch.optim.Adam,
        policy_opt_method: Type[torch.optim.Optimizer] = torch.optim.Adam,
        running_objective_type: str = "cost",
        critic_td_n: int = 1,
        cliprange: float = 0.2,
        discount_factor: float = 0.7,
        observer: Optional[Observer] = None,
        N_episodes: int = 2,
        N_iterations: int = 100,
        value_threshold: float = np.inf,
        stopping_criterion: Optional[Callable[[DataBuffer], bool]] = None,
        gae_lambda: float = 0.0,
        is_normalize_advantages: bool = True,
        device: str = "cpu",
        entropy_coeff: float = 0.0,
    ):
        """Initialize the object with the given parameters.

        Args:
            policy_model (PerceptronWithTruncatedNormalNoise): The
                neural network model parameterizing the policy.
            critic_model (ModelPerceptron): The neural network model
                used for the value function approximation.
            sampling_time (float): Time interval between two consecutive
                actions.
            running_objective (RunningObjective): A function that
                returns the scalar running cost or reward associated
                with an observation-action pair.
            simulator (Simulator): The simulation environment where the
                agent performs actions.
            critic_n_epochs (int): The number of epochs for which the
                critic is trained per iteration.
            policy_n_epochs (int): The number of epochs for which the
                policy is trained per iteration.
            critic_opt_method_kwargs (Dict[str, Any]): A dictionary of
                keyword arguments for the optimizer used for the critic.
            policy_opt_method_kwargs (Dict[str, Any]): A dictionary of
                keyword arguments for the optimizer used for the policy.
            critic_opt_method (Type[torch.optim.Optimizer]): The
                optimization algorithm class used for training the
                critic, e.g. torch.optim.Adam.
            policy_opt_method (Type[torch.optim.Optimizer]): The
                optimization algorithm class used for training the
                policy, e.g. torch.optim.Adam.
            running_objective_type (str): Specifies whether the running
                objective represents a 'cost' to minimize or a 'reward'
                to maximize.
            critic_td_n (int): The n-step temporal-difference parameter
                used for critic updates.
            epsilon (float): Clipping parameter that restricts the
                deviation of the new policy from the old policy.
            discount_factor (float): A factor applied to future rewards
                or costs to discount their value relative to immediate
                ones.
            observer (Optional[Observer]): Object responsible for state
                estimation from observations.
            N_episodes (int): The number of episodes to run in every
                iteration.
            N_iterations (int): The number of iterations to run in the
                scenario.
            value_threshold (float): Threshold of the value to
                end an episode.

        Raises:
            AssertionError: If the `running_objective_type` is invalid.

        Returns:
            None
        """
        assert (
            running_objective_type == "cost" or running_objective_type == "reward"
        ), f"Invalid 'running_objective_type' value: '{running_objective_type}'. It must be either 'cost' or 'reward'."
        super().__init__(
            **get_policy_gradient_kwargs(
                sampling_time=sampling_time,
                running_objective=running_objective,
                simulator=simulator,
                discount_factor=discount_factor,
                observer=observer,
                N_episodes=N_episodes,
                N_iterations=N_iterations,
                value_threshold=value_threshold,
                policy_type=PolicyPPO,
                policy_model=policy_model,
                policy_opt_method=policy_opt_method,
                policy_opt_method_kwargs=policy_opt_method_kwargs,
                is_reinstantiate_policy_optimizer=True,
                policy_kwargs=dict(
                    cliprange=cliprange,
                    running_objective_type=running_objective_type,
                    sampling_time=sampling_time,
                    gae_lambda=gae_lambda,
                    is_normalize_advantages=is_normalize_advantages,
                    entropy_coeff=entropy_coeff,
                ),
                policy_n_epochs=policy_n_epochs,
                critic_model=critic_model,
                critic_opt_method=critic_opt_method,
                critic_opt_method_kwargs=critic_opt_method_kwargs,
                critic_td_n=critic_td_n,
                critic_n_epochs=critic_n_epochs,
                critic_is_value_function=True,
                is_reinstantiate_critic_optimizer=True,
                stopping_criterion=stopping_criterion,
                device=device,
            )
        )


class SDPG(RLScenario):
    """Implements the Stochastic Deep Policy Gradient (SDPG) algorithm.

    SDPGScenario implements the Stochastic Deep Policy Gradient (SDPG) algorithm,
    an off-policy actor-critic method using a deep policy to model stochastic policies and
    a deep network as the critic that approximates a value Function.

    The SDPG algorithm performs policy updates by maximizing the critic's output and updates
    the critic using temporal-difference learning. This scenario orchestrates the training
    process, including interaction with the simulation environment, handling of the experience
    replay buffer, and coordination of policy and critic improvements.
    """

    def __init__(
        self,
        policy_model: PerceptronWithTruncatedNormalNoise,
        critic_model: ModelPerceptron,
        sampling_time: float,
        running_objective: RunningObjective,
        simulator: Simulator,
        critic_n_epochs: int,
        policy_n_epochs: int,
        critic_opt_method_kwargs: Dict[str, Any],
        policy_opt_method_kwargs: Dict[str, Any],
        critic_opt_method: Type[torch.optim.Optimizer] = torch.optim.Adam,
        policy_opt_method: Type[torch.optim.Optimizer] = torch.optim.Adam,
        critic_td_n: int = 1,
        discount_factor: float = 0.7,
        observer: Optional[Observer] = None,
        N_episodes: int = 2,
        N_iterations: int = 100,
        value_threshold: float = np.inf,
        stopping_criterion: Optional[Callable[[DataBuffer], bool]] = None,
        is_normalize_advantages=True,
        gae_lambda=0.0,
    ):
        """Initialize SDPG.

        Args:
            policy_model: The
                policy network model that defines the policy
                architecture.
            critic_model: The critic network model
                that defines the value function architecture.
            sampling_time (float): The time step between agent actions
                in the environment.
            running_objective (RunningObjective): Function calculating
                the reward or cost at each time step when an action is
                taken.
            simulator (Simulator): The environment in which the agent
                operates, providing observation.
            critic_n_epochs (int): Number of epochs for which the critic
                network is trained at each iteration.
            policy_n_epochs (int): Number of epochs for which the policy
                network is trained at each iteration.
            critic_opt_method_kwargs (Dict[str, Any]): Parameters for
                the critic's optimization algorithm.
            policy_opt_method_kwargs (Dict[str, Any]): Parameters for
                the policy's optimization algorithm.
            critic_opt_method (Type[torch.optim.Optimizer]): Class of
                the optimizer to use for optimizing the critic,
                defaults to torch.optim.Adam
            policy_opt_method (Type[torch.optim.Optimizer]): Class of
                the optimizer to use for optimizing the policy, ,
                defaults to torch.optim.Adam
            critic_td_n (int): The number of steps to look ahead in the
                TD-target for the critic.
            discount_factor (float): The factor by which future rewards
                or costs are discounted relative to immediate running
                objectives.
            observer (Optional[Observer]): The observer object that
                estimates the state of the environment from
                observations.
            N_episodes (int): The number of episodes per iteration.
            N_iterations (int): The total number of iterations for
                training.
            value_threshold (float): Threshold for the value
                that, once reached, stops the episode, defaults to
                np.inf.
        """
        super().__init__(
            **get_policy_gradient_kwargs(
                sampling_time=sampling_time,
                running_objective=running_objective,
                simulator=simulator,
                discount_factor=discount_factor,
                observer=observer,
                N_episodes=N_episodes,
                N_iterations=N_iterations,
                value_threshold=value_threshold,
                policy_type=PolicySDPG,
                policy_model=policy_model,
                policy_opt_method=policy_opt_method,
                policy_opt_method_kwargs=policy_opt_method_kwargs,
                is_reinstantiate_policy_optimizer=False,
                policy_n_epochs=policy_n_epochs,
                critic_model=critic_model,
                critic_opt_method=critic_opt_method,
                critic_opt_method_kwargs=critic_opt_method_kwargs,
                critic_td_n=critic_td_n,
                critic_n_epochs=critic_n_epochs,
                critic_is_value_function=True,
                is_reinstantiate_critic_optimizer=True,
                policy_kwargs=dict(
                    sampling_time=sampling_time,
                    is_normalize_advantages=is_normalize_advantages,
                    gae_lambda=gae_lambda,
                ),
                stopping_criterion=stopping_criterion,
            )
        )


class REINFORCE(RLScenario):
    """Implements the REINFORCE algorithm."""

    def __init__(
        self,
        policy_model: PerceptronWithTruncatedNormalNoise,
        sampling_time: float,
        running_objective: RunningObjective,
        simulator: Simulator,
        policy_opt_method_kwargs: Dict[str, Any],
        policy_opt_method: Type[torch.optim.Optimizer] = torch.optim.Adam,
        policy_n_epochs: int = 1,
        discount_factor: float = 1.0,
        observer: Optional[Observer] = None,
        N_episodes: int = 4,
        N_iterations: int = 100,
        value_threshold: float = np.inf,
        is_with_baseline: bool = True,
        is_do_not_let_the_past_distract_you: bool = True,
        stopping_criterion: Optional[Callable[[DataBuffer], bool]] = None,
    ):
        """Initialize an REINFORCE object.

        Args:
            policy_model: The
                policy network model that defines the policy
                architecture.
            sampling_time: The time step between agent actions
                in the environment.
            running_objective: Function calculating
                the reward or cost at each time step when an action is
                taken.
            simulator: The environment in which the agent
                operates, providing state, observation.
            policy_opt_method_kwargs: The keyword
                arguments for the policy optimizer method.
            policy_opt_method (Type[torch.optim.Optimizer], optional):
                The policy optimizer method.
            n_epochs: The number of epochs used by the
                policy optimizer.
            discount_factor: The discount factor used
                by the RLScenario. Defaults to 1.0.
            observer: The observer object
                that estimates the state of the environment from observations.
            N_episodes: The number of episodes per iteration.
            N_iterations: The total number of iterations
                for training.
            is_with_baseline: Whether to use baseline as value (i.e. cumulative cost or reward)
                from previous iteration.
            is_do_not_let_the_past_distract_you: Whether to use tail total costs or not.

        Returns:
            None: None
        """
        super().__init__(
            **get_policy_gradient_kwargs(
                sampling_time=sampling_time,
                running_objective=running_objective,
                simulator=simulator,
                discount_factor=discount_factor,
                observer=observer,
                N_episodes=N_episodes,
                N_iterations=N_iterations,
                value_threshold=value_threshold,
                policy_type=PolicyReinforce,
                policy_model=policy_model,
                policy_n_epochs=policy_n_epochs,
                policy_opt_method=policy_opt_method,
                policy_opt_method_kwargs=policy_opt_method_kwargs,
                is_reinstantiate_policy_optimizer=False,
                policy_kwargs=dict(
                    is_with_baseline=is_with_baseline,
                    is_do_not_let_the_past_distract_you=is_do_not_let_the_past_distract_you,
                ),
                is_use_critic_as_policy_kwarg=False,
                stopping_criterion=stopping_criterion,
            ),
        )


class DDPG(RLScenario):
    """Implements a scenario for interacting with an environment using the Deep Deterministic Policy Gradients (DDPG) algorithm."""

    def __init__(
        self,
        policy_model: PerceptronWithTruncatedNormalNoise,
        critic_model: ModelPerceptron,
        sampling_time: float,
        running_objective: RunningObjective,
        simulator: Simulator,
        critic_n_epochs: int,
        policy_n_epochs: int,
        critic_opt_method_kwargs: Dict[str, Any],
        policy_opt_method_kwargs: Dict[str, Any],
        critic_opt_method: Type[torch.optim.Optimizer] = torch.optim.Adam,
        policy_opt_method: Type[torch.optim.Optimizer] = torch.optim.Adam,
        critic_td_n: int = 1,
        discount_factor: float = 0.7,
        observer: Optional[Observer] = None,
        N_episodes: int = 2,
        N_iterations: int = 100,
        value_threshold: float = np.inf,
        stopping_criterion: Optional[Callable[[DataBuffer], bool]] = None,
    ):
        """Instantiate DDPG Scenario.

        Args:
            policy_model (PerceptronWithTruncatedNormalNoise): The
                policy (actor) neural network model with input as state
                and output as action.
            critic_model (ModelPerceptron): The critic neural network
                model with input as state-action pair and output as
                Q-value estimate.
            sampling_time (float): The time interval between each action
                taken by the policy.
            running_objective (RunningObjective): Th function
                calculating the running cost at each time step when an
                action is taken.
            simulator (Simulator): The environment simulator in which
                the agent operates.
            critic_n_epochs (int): The number of epochs for training the
                critic model during each optimization.
            policy_n_epochs (int): The number of epochs for training the
                policy model during each optimization.
            critic_opt_method_kwargs (Dict[str, Any]): Keyword arguments
                for the critic optimizer method.
            policy_opt_method_kwargs (Dict[str, Any]): Keyword arguments
                for the policy optimizer method.
            critic_opt_method (Type[torch.optim.Optimizer]): The
                optimizer class to be used for the critic. Defaults to
                torch.nn.Adam.
            policy_opt_method (Type[torch.optim.Optimizer]): The
                optimizer class to be used for the policy. Defaults to
                torch.nn.Adam.
            critic_td_n (int): The n-step return for temporal-difference
                learning for the critic estimator.
            discount_factor (float): The discount factor that weighs
                future costs lower compared to immediate costs.
            observer (Optional[Observer]): An observer object used for
                deriving state estimations from raw observations.
            N_episodes (int): The number of episodes to be executed in
                each training iteration.
            N_iterations (int): The total number of training iterations
                for the scenario.
            value_threshold (float): The threshold for the total
                cumulative objective value that triggers the end of an
                episode.
        """
        super().__init__(
            **get_policy_gradient_kwargs(
                sampling_time=sampling_time,
                running_objective=running_objective,
                simulator=simulator,
                discount_factor=discount_factor,
                observer=observer,
                N_episodes=N_episodes,
                N_iterations=N_iterations,
                value_threshold=value_threshold,
                policy_type=PolicyDDPG,
                policy_model=policy_model,
                policy_opt_method=policy_opt_method,
                policy_opt_method_kwargs=policy_opt_method_kwargs,
                is_reinstantiate_policy_optimizer=True,
                policy_n_epochs=policy_n_epochs,
                critic_model=critic_model,
                critic_opt_method=critic_opt_method,
                critic_opt_method_kwargs=critic_opt_method_kwargs,
                critic_td_n=critic_td_n,
                critic_n_epochs=critic_n_epochs,
                critic_is_value_function=False,
                is_reinstantiate_critic_optimizer=True,
                stopping_criterion=stopping_criterion,
            )
        )


# Additional classes such as Policy may need to be defined or imported depending on the setup of your framework.



"""A flexibly configurable framework for agent-enviroment simulation with a menu of predictive and safe reinforcement learning scenarios.

It is made for researchers and engineers in reinforcement learning and control theory.
"""

__version__ = "0.3.0"

import argparse
import datetime
import glob
import platform
import shelve
import subprocess
import sys
import os
import inspect
import logging
import traceback
import types
import warnings

import matplotlib

matplotlib.use("Agg")

import random
from click import Option
from fsspec import Callback
from matplotlib import interactive
import omegaconf
import rich

from omegaconf import DictConfig, OmegaConf, ListConfig
from omegaconf.resolvers import oc
import re

from typing import Any

from omegaconf import Container
from omegaconf._utils import _DEFAULT_MARKER_
from omegaconf.errors import ConfigKeyError, InterpolationResolutionError


import hashlib

from omegaconf import DictConfig
import rehydra
from rehydra._internal.utils import _locate

import rehydra.core.plugins
import rehydra._internal.config_loader_impl
from scipy.optimize import OptimizeWarning
from rehydra.utils import instantiate as inst


import mlflow

from unittest.mock import Mock


import plotly.graph_objects as go
import json


import tempfile

from pathlib import Path

import numpy

from unittest.mock import MagicMock


from rehydra import main as rehydramain
from . import __internal
from . import callback
from regelum.__internal.base import RegelumBase
from .__internal.metadata import Metadata
from .animation import StateAnimation

from .callback import (
    OnEpisodeDoneCallback,
    OnIterationDoneCallback,
    ConfigDiagramCallback,
    TimeRemainingCallback,
    SaveProgressCallback,
)
import time
import pandas as pd

from .configure import get_user_settings, config_file

# try:
#     import torch
# except (ModuleNotFoundError, ImportError):
#     torch = MagicMock()

import torch

from . import model
from . import data_buffers
from .optimizable import *
from . import critic
from . import scenario
from . import policy
from . import observer
from .event import Event

from rich import print, pretty
from rich.panel import Panel
import rich.logging
from rich.logging import RichHandler

mock = Mock()


ANIMATION_TYPES_SAVE_FORMATS = ["html", "mp4"]
ANIMATION_TYPES_REQUIRING_SAVING_SCENARIO_PLAYBACK = [
    "playback"
] + ANIMATION_TYPES_SAVE_FORMATS
ANIMATION_TYPES_NONE = [None, "None"]
ANIMATION_TYPES_REQUIRING_ANIMATOR = [
    "live",
] + ANIMATION_TYPES_REQUIRING_SAVING_SCENARIO_PLAYBACK
ANIMATION_TYPES = ANIMATION_TYPES_NONE + ANIMATION_TYPES_REQUIRING_ANIMATOR


if "REGELUM_DATA_DIR" not in os.environ:
    os.environ["REGELUM_DATA_DIR"] = os.path.abspath(sys.path[0]) + "/regelum_data"
else:
    os.environ["REGELUM_DATA_DIR"] = os.path.abspath(os.environ["REGELUM_DATA_DIR"])

if "MLFLOW_TRACKING_URI" not in os.environ:
    os.environ["MLFLOW_TRACKING_URI"] = (
        f"file://{os.environ['REGELUM_DATA_DIR']}/mlruns"
    )


def hash_string(s):
    return int(hashlib.sha1(s.encode("utf-8")).hexdigest(), base=16)


import subprocess, os, platform


def start(filepath):
    if platform.system() == "Darwin":  # macOS
        subprocess.run(["open", filepath], check=True)
    elif platform.system() == "Windows":  # Windows
        os.startfile(filepath)
    else:  # linux variants
        subprocess.run(["xdg-open", filepath], check=True)


def __memorize_instance(resolver):
    def inner(
        key: str,
        default: Any = _DEFAULT_MARKER_,
        *,
        _parent_: Container,
    ):
        if default == _DEFAULT_MARKER_:
            default = key.strip()
        instance_name = str(default)
        if instance_name in main.objects_created:
            return main.objects_created[instance_name]
        else:
            obj = inst(resolver(key, default=default, _parent_=_parent_), path=default)
            main.objects_created[instance_name] = obj
            return obj

    return inner


def __sub_map(pattern, f, s):
    def map_match(match):
        return f(match.group())

    return re.sub(pattern, map_match, s)


def obtain(obj_repr):
    if not isinstance(obj_repr, str):
        obj_repr = str(obj_repr)
    obj_repr = (
        obj_repr.replace("__QUOTATION__", '"')
        .replace("__APOSTROPHE__", "'")
        .replace("__TILDE__", "~")
    )
    obj_repr = rehydra._internal.core_plugins.file_config_source.numerize_string(
        obj_repr
    )
    pattern = re.compile(r"(\A|[^a-zA-Z0-9\._])[a-zA-Z_][a-zA-Z0-9_]*")
    resolved = []

    def resolve(s):
        if s[0].isalnum() or s[0] == "_":
            prefix = ""
        else:
            prefix = s[0]
            s = s[1:]
        try:
            if s == "np":
                s = "numpy"
            entity = _locate(s)
        except Exception:
            entity = eval(s)
        resolved.append(entity)
        return f"{prefix}resolved[{len(resolved) - 1}]"

    obj_repr_resolved_modules = __sub_map(pattern, resolve, obj_repr)
    return eval(obj_repr_resolved_modules)


OmegaConf.register_new_resolver("same", __memorize_instance(oc.select))
OmegaConf.register_new_resolver(name="get", resolver=obtain)
OmegaConf.register_new_resolver(name="mock", resolver=lambda: mock)


class ComplementedConfig:
    """A config object, generated by ``regelum``'s config scenario."""

    def __init__(
        self, cfg: Union[DictConfig, ListConfig], config_path: str = ""
    ) -> None:
        """Initialize a complemented config.

        Args:
            cfg: rehydra config object
            config_path: path to a config file
        """
        self._rehydra_config = cfg
        self.config_path = config_path
        self._saved_hash = None

    def refresh(self):
        """Reset all cached instances.

        Call this if you made changes to your config dynamically and would like those change to take
        effect in forthcoming instantiations.
        """
        main.objects_created.clear()

    def __str__(self):
        return (
            str(self._rehydra_config)
            .replace("DictConfig", "ComplementedConfigDict")
            .replace("ListConfig", "ComplementedConfigList")
            .replace("${get:", "={")
            .replace("${same:", "~{")
        )

    def __repr__(self):
        return self.__class__.__name__ + "( " + repr(self._rehydra_config) + ")"

    def __hash__(self):
        if not self._saved_hash:
            real_config = omegaconf.OmegaConf.to_container(self._rehydra_config)
            return hash_string(json.dumps(real_config, sort_keys=True))
        else:
            return self._saved_hash

    def copy(self):
        return self.__class__(self._rehydra_config.copy(), config_path=self.config_path)

    def __invert__(self):
        """Instantiate the object described by the config (recursively).

        ``_target_`` has to be specified.
        """
        # if not self.config_path:
        return inst(self._rehydra_config, path=self.config_path)
        # else:
        #     instance_name = self.config_path.strip()
        #     if instance_name in self.objects_created:
        #         return self.objects_created[instance_name]
        #     else:
        #         res = inst(self.__rehydra_config, path=self.config_path)
        #         self.objects_created[instance_name] = res
        #         return res


class ComplementedConfigList(ComplementedConfig):
    """A config object, generated by ``regelum``'s config scenario that corresponds to a complemented version of `ListConfig` from omegaconf."""

    def __contains__(self, item):
        return item in self._rehydra_config

    def __getitem__(self, key):
        cfg = object.__getattribute__(self, "_rehydra_config")
        name = key
        item = cfg[name]
        if isinstance(item, DictConfig) or isinstance(item, ListConfig):
            if self.config_path:
                child_config_path = f"{self.config_path}[{name}]"
            else:
                raise ValueError("Top level list configs are not supported.")
            entity = (
                ComplementedConfigList
                if isinstance(item, ListConfig)
                else ComplementedConfigDict
            )
            item = entity(item, config_path=child_config_path)
        return item

    def __setitem__(self, key, value):
        self._rehydra_config[key] = value

    def __delitem__(self, key):
        del self._rehydra_config[key]


class ComplementedConfigDict(ComplementedConfig):
    """A config object, generated by ``regelum``'s config scenario that corresponds to a complemented version of `DictConfig` from omegaconf."""

    def treemap(self, root="config"):
        def format(
            parent_name, parent_node, parent_node_raw, occupied=None, parent_color=50
        ):
            if occupied is None:
                occupied = set()
            labels = []
            parents = []
            colors = []
            text = []
            if isinstance(parent_node, ComplementedConfigDict):
                items = parent_node.items()
            else:
                items = enumerate(parent_node)
            for name, value in items:
                # check if node as attribute value
                parents.append(parent_name)
                # if type(parent_node_raw) is str:
                #    parent_node_raw = omegaconf.OmegaConf.to_container(parent_node.__rehydra_config)
                if not isinstance(name, int):
                    real_name = (
                        name if name in parent_node_raw else name + "__IGNORE__"
                    ).strip()
                else:
                    real_name = name
                    name = str(name)
                while name in occupied:
                    name += " "
                if isinstance(value, ComplementedConfig):
                    node_raw = parent_node_raw[real_name]
                    if type(node_raw) is str:
                        text.append(node_raw.replace("__IGNORE__", "%%"))
                        node_raw = omegaconf.OmegaConf.to_container(
                            value._rehydra_config
                        )
                    else:
                        text.append("")
                    labels.append(name)
                    colors.append((parent_color * 1.2) % 100)
                    (
                        subnode_parents,
                        subnode_labels,
                        subnode_colors,
                        subnode_text,
                    ) = format(
                        name,
                        value,
                        node_raw,
                        occupied=occupied,
                        parent_color=(parent_color * 1.2) % 100,
                    )
                    for i, subnode_label in enumerate(subnode_labels):
                        if subnode_label in labels or subnode_label in parents:
                            subnode_labels[i] = subnode_labels[i] + " "
                    labels += subnode_labels
                    parents += subnode_parents
                    colors += subnode_colors
                    text += subnode_text
                else:
                    if (
                        type(parent_node_raw[real_name]) is str
                        and "$" in parent_node_raw[real_name]
                    ):
                        text.append(
                            parent_node_raw[real_name].replace("__IGNORE__", "%%")
                        )
                    else:
                        text.append("")
                    colors.append(hash_string(name) % 100)
                    name = f"{name}: {str(value)}"
                    while name in occupied:
                        name += " "
                    labels.append(name)

                occupied.add(labels[-1])
            occupied.add(parent_name)
            return parents, labels, colors, text

        raw_self = omegaconf.OmegaConf.to_container(self._rehydra_config)
        self._saved_hash = hash_string(json.dumps(raw_self, sort_keys=True))
        parents, labels, colors, text = format(
            f"{root} {hex(hash(self))}", self, raw_self
        )
        # parents = [parent[:-1] if "_" in parent else parent for parent in parents]
        # parents = [""] + parents
        # labels = ["config"] + labels
        fig = go.Figure(
            go.Treemap(
                labels=labels,
                parents=parents,
                marker=dict(colors=colors, colorscale="RdBu", cmid=50),
                text=text,
            )
        )
        return fig

    def __contains__(self, item):
        return (
            item in self._rehydra_config or item + "__IGNORE__" in self._rehydra_config
        )

    def __getattr__(self, item):
        cfg = object.__getattribute__(self, "_rehydra_config")
        try:
            name = item
            attr = cfg.__getattr__(name)
        except (IndexError, KeyError, AttributeError, ConfigKeyError):
            name = item + "__IGNORE__"
            attr = cfg.__getattr__(name)
        if isinstance(attr, DictConfig) or isinstance(attr, ListConfig):
            if self.config_path:
                child_config_path = self.config_path + "." + name
            else:
                child_config_path = name
            entity = (
                ComplementedConfigList
                if isinstance(item, ListConfig)
                else ComplementedConfigDict
            )
            attr = entity(attr, config_path=child_config_path)
        return attr

    def __getitem__(self, key):
        cfg = object.__getattribute__(self, "_rehydra_config")
        try:
            name = key
            item = cfg[name]
        except (IndexError, KeyError, AttributeError, ConfigKeyError) as e:
            if isinstance(key, str):
                name = key + "__IGNORE__"
                item = cfg[name]
            else:
                raise e
        if isinstance(item, DictConfig) or isinstance(item, ListConfig):
            if self.config_path:
                child_config_path = self.config_path + "." + name
            else:
                child_config_path = name
            entity = (
                ComplementedConfigList
                if isinstance(item, ListConfig)
                else ComplementedConfigDict
            )
            item = entity(item, config_path=child_config_path)
        return item

    def __setitem__(self, key, value):
        if key + "__IGNORE__" in self._rehydra_config:
            self._rehydra_config[key + "__IGNORE__"] = value
        else:
            self._rehydra_config[key] = value

    def __setattr__(self, key, value):
        if key in ["config_path", "_saved_hash"]:
            object.__setattr__(self, key, value)
            return
        if hasattr(self, "_rehydra_config"):
            if key + "__IGNORE__" in self._rehydra_config:
                self._rehydra_config.__setattr__(key + "__IGNORE__", value)
            else:
                self._rehydra_config.__setattr__(key, value)
        else:
            object.__setattr__(self, key, value)

    def has_key(self, key):
        return key in self._rehydra_config or key + "__IGNORE__" in self._rehydra_config

    def clear(self):
        self._rehydra_config.clear()

    def update(self, *args, **kwargs):
        for subdict in args:
            for key, value in subdict.values():
                self[key] = value
        for key, value in kwargs.values():
            self[key] = value

    def keys(self):
        return [key.replace("__IGNORE__", "") for key in self._rehydra_config.keys()]

    def items(self):
        return [
            (
                key.replace("__IGNORE__", ""),
                (
                    value
                    if not isinstance(value, DictConfig)
                    and not isinstance(value, ListConfig)
                    else (
                        ComplementedConfigDict
                        if isinstance(value, DictConfig)
                        else ComplementedConfigList
                    )(
                        value,
                        config_path=(
                            key if not self.config_path else f"{self.config_path}.{key}"
                        ),
                    )
                ),
            )
            for key, value in self._rehydra_config.items()
        ]

    def values(self):
        return [
            (
                value
                if not isinstance(value, DictConfig)
                and not isinstance(value, ListConfig)
                else (
                    ComplementedConfigDict
                    if isinstance(value, DictConfig)
                    else ComplementedConfigList
                )(
                    value,
                    config_path=(
                        key if not self.config_path else f"{self.config_path}.{key}"
                    ),
                )
            )
            for key, value in self._rehydra_config.items()
        ]

    def __delitem__(self, key):
        if key + "__IGNORE__" in self._rehydra_config:
            del self._rehydra_config[key + "__IGNORE__"]
        else:
            del self._rehydra_config[key]


class RegelumExitException(Exception):
    """Raised to forcefully shut down current simulation."""

    pass


logging.basicConfig(
    level="INFO", format="%(message)s", datefmt="[%X]", handlers=[RichHandler()]
)

pretty.install()


logo_ascii = """██████╗ ███████╗ ██████╗ ███████╗██╗     ██╗   ██╗███╗   ███╗
██╔══██╗██╔════╝██╔════╝ ██╔════╝██║     ██║   ██║████╗ ████║
██████╔╝█████╗  ██║  ███╗█████╗  ██║     ██║   ██║██╔████╔██║
██╔══██╗██╔══╝  ██║   ██║██╔══╝  ██║     ██║   ██║██║╚██╔╝██║
██║  ██║███████╗╚██████╔╝███████╗███████╗╚██████╔╝██║ ╚═╝ ██║
╚═╝  ╚═╝╚══════╝ ╚═════╝ ╚══════╝╚══════╝ ╚═════╝ ╚═╝     ╚═╝"""


class main:
    """The decorator used to invoke ``regelum``'s config scenario.

    Use it to decorate your ``main`` function like so:
    ::

         @regelum.main(config_path=..., config_name=...)
         def my_main(config):
             ...

         if __name__ == '__main__':
             my_main()
    """

    callbacks = None
    logger = None
    metadata = None
    config = None
    assignments = []
    weak_assignments = []
    builtin_callbacks = [
        OnEpisodeDoneCallback,
        OnIterationDoneCallback,
        ConfigDiagramCallback,
        TimeRemainingCallback,
        SaveProgressCallback,
    ]
    objects_created = {}

    class RegelumArgumentParser(argparse.ArgumentParser):
        """A parser object designed for handling peculiar nuances of arg parsing while interfacing with rehydra."""

        def __init__(self, *args, **kwargs):
            """Initialize and instance of RegelumArgumentParser.

            :param args: arguments to pass to base
            :param kwargs: keyword arguments to pass to base
            """
            self._stored_args = []
            self._registered_args = []
            self._triggers = {}
            super().__init__(*args, **kwargs)

        def add_argument(self, *args, trigger=None, **kwargs):
            res = super().add_argument(*args, **kwargs)
            self._registered_args = self._registered_args + list(args)
            for arg in args:
                if arg.startswith("--"):
                    key = arg[2:]
                elif arg.startswith("-"):
                    key = arg[1:]
                else:
                    raise ValueError(
                        "Argument name should start with either a '-' or a '--'."
                    )
                key = key.replace("-", "_")
                if key == "help" or key == "h":
                    continue
                self._triggers[key] = trigger if trigger is not None else lambda v: None
            return res

        def parse_args(self):
            if "--single-thread" in sys.argv:
                for i, arg in enumerate(sys.argv):
                    if "--jobs" in arg or "launcher.n_jobs" in arg:
                        print(f"Ingnoring {arg}, since --single-thread was specified.")
                        sys.argv.pop(i)
                        break
            self.old_args = sys.argv
            self._stored_args = [
                arg for arg in sys.argv if arg.split("=")[0] in self._registered_args
            ]
            res = super().parse_args(self._stored_args)
            sys.argv = [arg for arg in sys.argv if arg not in self._stored_args]
            for key in self._triggers:
                self._triggers[key](vars(res)[key])
            self._registered_args = []

            sys.argv.insert(1, "--multirun")
            sys.argv.append("rehydra.job.chdir=True")

            return res

        def restore_args(self):
            sys.argv = self.old_args

    @classmethod
    def post_weak_assignment(cls, key, value):
        cls.weak_assignments.append((key, value))

    @classmethod
    def post_assignment(cls, key, value, weak=False):
        if weak:
            cls.post_weak_assignment(key, value)
        else:
            cls.assignments.append((key, value))

    @classmethod
    def apply_assignments(cls, cfg):
        for key, value in cls.weak_assignments:
            if callable(value):
                value = value(cfg)
            current = eval(f"cfg.{key.replace('%%', '')}")
            if current == "__REPLACE__":
                exec(f"cfg.{key.replace('%%', '')} = value")
        for key, value in cls.assignments:
            if callable(value):
                value = value(cfg)
            exec(f"cfg.{key.replace('%%', '')} = value")

    def __init__(
        self,
        *args,
        logger=None,
        callbacks=None,
        **kwargs,
    ):
        """Create an instance of ``regelum.main``.

        :param config_path: path to the folder containing your config(s)
        :type config_path: str
        :param config_name: the name of your config (without extension)
        :type config_name: str
        :param callbacks: list of uninstantiated callbacks (classes) to be registered
        :type callbacks: list[class], optional
        :param logger: a logger instance to be supplied to callbacks
        :type logger: Logger, optional

        """
        print(
            Panel.fit(
                logo_ascii,
                subtitle=f"v{__version__}",
                title="AIDA Solutions, 2024",
                style="magenta",
            )
        )
        if logger is None:
            logger = logging.getLogger("regelum")
            # logger.info("Try")
        if callbacks is None:
            callbacks = []

        if "--configure" in sys.argv:
            if os.path.exists(config_file):
                os.remove(config_file)
            sys.argv.pop(sys.argv.index("--configure"))

        user_settings = get_user_settings()
        for name in user_settings:
            if name not in os.environ:
                if user_settings[name] is not None:
                    os.environ[name] = str(user_settings[name])

        self.callbacks_ = callbacks + self.builtin_callbacks

        self.parser = self.RegelumArgumentParser(description="Run with config ...")

        self.parser.add_argument("--no-git", action="store_true")

        def disable_logging(flag):
            if flag:
                sys.argv.append("rehydra/job_logging=disabled")

        self.parser.add_argument(
            "--disable-logging", action="store_true", trigger=disable_logging
        )

        def disable_callbacks(flag):
            if flag:
                self.__class__.builtin_callbacks = []
                self.callbacks_ = []

        self.parser.add_argument(
            "--disable-callbacks", action="store_true", trigger=disable_callbacks
        )

        self.parser.add_argument("--enable-streamlit", action="store_true")

        self.parser.add_argument("--interactive", action="store_true")
        self.parser.add_argument("--show-plots", action="store_true")
        self.parser.add_argument("--playback", action="store_true")

        self.parser.add_argument("--fps", default=2.5)

        self.parser.add_argument("--cooldown-factor", default=1.0)

        self.parser.add_argument("--save-animation", action="store_true")
        self.parser.add_argument("--parallel", action="store_true")

        def single_thread(flag):
            if not flag:
                sys.argv.append("rehydra/launcher=joblib")

        self.parser.add_argument(
            "--single-thread", action="store_true", trigger=single_thread
        )

        def sweep(flag):
            if flag:
                sys.argv.append("rehydra/sweeper=ax")
                self.is_sweep = True
            else:
                self.is_sweep = False

        self.parser.add_argument("--sweep", action="store_true")

        def tags(val):
            if val:
                tags_str = val.split(",")
                self.tags = {tag.split(":")[0]: tag.split(":")[1] for tag in tags_str}
            else:
                self.tags = {}

        self.parser.add_argument("--tags", trigger=tags)

        def experiment(val):
            if val:
                self.experiment_name = val

        self.parser.add_argument("--experiment", trigger=experiment)

        def jobs(val):
            if val is not None:
                sys.argv.append(f"rehydra.launcher.n_jobs={val}")

        self.parser.add_argument("--jobs", trigger=jobs)

        self.parser.add_argument("--skip-frames", action="store", default=1, type=int)

        self.tags = {}
        self.experiment_name = "Default"

        self.__class__.is_clear_matplotlib_cache_in_callbacks = True
        for arg in sys.argv:
            if "live" in arg:
                self.__class__.is_clear_matplotlib_cache_in_callbacks = False

        self.args = args
        self.kwargs = kwargs
        self.kwargs["version_base"] = (
            None if "version_base" not in kwargs else kwargs["version_base"]
        )
        self.kwargs["config_path"] = (
            "." if "config_path" not in kwargs else kwargs["config_path"]
        )
        self.__class__.logger = logger

    def __call__(self, old_app):
        def rcognita_main(*args, **kwargs):

            argv = self.parser.parse_args()
            common_dir = tempfile.TemporaryDirectory()
            initial_working_directory = os.getcwd()
            initial_pythonpath = (
                os.environ["PYTHONPATH"] if "PYTHONPATH" in os.environ else ""
            )
            script_path = inspect.getsourcefile(old_app)
            no_git = argv.no_git
            skip_frames = argv.skip_frames
            path_main = os.path.abspath(script_path)
            path_parent = "/".join(path_main.split("/")[:-1])
            os.chdir(path_parent)
            path = os.path.abspath(self.kwargs["config_path"])
            self.kwargs["config_path"] = path

            def app(
                cfg,
                argv=argv,
                logger=self.__class__.logger,
                is_clear_matplotlib_cache_in_callbacks=self.__class__.is_clear_matplotlib_cache_in_callbacks,
            ):
                # args = self.parser.parse_args()
                if "seed" in cfg:
                    seed = cfg["seed"]
                    delattr(cfg, "seed")
                else:
                    seed = 0

                mlflow.set_tracking_uri(os.environ["MLFLOW_TRACKING_URI"])
                numpy.random.seed(seed)
                torch.manual_seed(seed)
                random.seed(seed)
                with shelve.open(".report") as f:
                    f["termination"] = "running"
                    f["started"] = datetime.datetime.now()
                    f["args"] = sys.argv
                    f["config_path"] = path
                    f["script_path"] = script_path
                    f["PYTHONPATH"] = initial_pythonpath
                    f["initial_cwd"] = initial_working_directory
                    f["common_dir"] = common_dir.name
                    f["hostname"] = platform.node()
                    f["seed"] = seed
                os.mkdir("gfx")
                os.mkdir(".callbacks")

                with omegaconf.flag_override(cfg, "allow_objects", True):
                    self.__class__.metadata = {
                        "no_git": no_git,
                        "skip_frames": skip_frames,
                        "logger": logger,
                        "script_path": script_path,
                        "config_path": path + f"/{self.kwargs['config_name']}.yaml",
                        "initial_working_directory": initial_working_directory,
                        "initial_pythonpath": initial_pythonpath,
                        "common_dir": common_dir.name,
                        "id": os.getcwd().split("/")[-1],
                        "report": lambda: shelve.open(
                            common_dir.name
                            + "/report_"
                            + os.getcwd().split("/")[-1].zfill(5)
                        ),
                        "pid": os.getpid(),
                        "argv": argv,
                        "main": self,
                    }
                    with Metadata(self.__class__.metadata):
                        self.callbacks = [callback_() for callback_ in self.callbacks_]
                        self.__class__.is_clear_matplotlib_cache_in_callbacks = (
                            is_clear_matplotlib_cache_in_callbacks
                        )
                        self.__class__.logger = logger
                        ccfg = ComplementedConfigDict(cfg)
                        self.apply_assignments(ccfg)
                        if "callbacks" in cfg and not argv.disable_callbacks:
                            for callback_ in cfg.callbacks:
                                callback_ = (
                                    obtain(callback_)
                                    if isinstance(callback_, str)
                                    else callback_
                                )
                                self.callbacks.append(callback_())
                            delattr(cfg, "callbacks")
                        elif "callbacks" in cfg:
                            delattr(cfg, "callbacks")
                        self.__class__.config = ccfg

                        try:
                            for callback_ in self.callbacks:
                                if callback_.cooldown:
                                    callback_.cooldown *= argv.cooldown_factor
                                callback_.on_launch()  # make sure this line is adequate to mlflow functionality
                            with self.__class__.metadata["report"]() as r:
                                r["path"] = os.getcwd()
                                r["pid"] = os.getpid()
                            self.tags.update({"run_path": os.getcwd()})

                            experiment_id = mlflow.set_experiment(
                                self.experiment_name
                            ).experiment_id

                            with mlflow.start_run(
                                experiment_id=experiment_id,
                                tags=self.tags,
                                run_name=" ".join(os.getcwd().split("/")[-3:]),
                            ):
                                overrides = {
                                    line.split("=")[0]
                                    .replace("+", "")
                                    .replace("/", "-"): line.split("=")[1]
                                    for line in OmegaConf.load(
                                        ".rehydra/overrides.yaml"
                                    )
                                }
                                mlflow.log_params(overrides)
                                mlflow.log_artifact("SUMMARY.html")
                                mlflow.log_artifact(".summary")
                                mlflow.log_artifact(".rehydra")
                                res = old_app(ccfg)
                                mlflow.log_artifact(".callbacks")
                                try:
                                    mlflow.log_artifact("__init__.log")
                                except FileNotFoundError:
                                    mlflow.log_artifact(
                                        "conftest.log"
                                    )  ## TO DO: Find a better way to handle this
                                if os.path.exists("callbacks.dill"):
                                    mlflow.log_artifact("callbacks.dill")

                            self.callbacks[0].log("Script terminated successfully.")
                        except RegelumExitException as e:
                            res = e
                        except InterpolationResolutionError as e:
                            with self.__class__.metadata["report"]() as r:
                                r["traceback"] = traceback.format_exc()
                            res = e
                            self.callbacks[0].log(
                                "Script terminated with error. This error occurred when trying to instantiate something from the config."
                            )
                            self.callbacks[0].exception(e)
                        except Exception as e:
                            with self.__class__.metadata["report"]() as r:
                                r["traceback"] = traceback.format_exc()
                            res = e
                            self.callbacks[0].log("Script terminated with error.")
                            self.callbacks[0].exception(e)
                        for callback_ in self.callbacks:
                            try:
                                callback_.on_termination(res)
                            except Exception as e:
                                callback_.log(
                                    f"Termination procedure for {callback_.__class__.__name__} failed."
                                )
                                callback_.exception(e)
                        with shelve.open(".report") as f:
                            f["termination"] = (
                                "successful"
                                if not isinstance(res, Exception)
                                else "".join(traceback.format_tb(res.__traceback__))
                            )
                            f["finished"] = datetime.datetime.now()
                        ccfg.refresh()
                    if argv.sweep:
                        return res
                    else:
                        if argv.playback:
                            filenames = glob.glob(".callbacks/**/*.html")
                            n_animations = len(
                                set([Path(filename).parent for filename in filenames])
                            )
                            latest_files = sorted(filenames, key=os.path.getctime)
                            for filename in latest_files[-n_animations:]:
                                start(filename)
                        if argv.show_plots:
                            start("SUMMARY.html")
                        return {
                            "result": res,
                            "callbacks": self.callbacks,
                            "directory": os.getcwd(),
                        }

            app.__module__ = old_app.__module__
            res = rehydramain(*self.args, **self.kwargs)(app)(*args, **kwargs)
            common_dir.cleanup()
            self.parser.restore_args()
            return res

        return rcognita_main


warnings.filterwarnings("ignore", category=UserWarning, module=rehydra.__name__)
warnings.filterwarnings("ignore", category=pd.io.pytables.PerformanceWarning)
warnings.filterwarnings("ignore", category=OptimizeWarning)
warnings.filterwarnings("ignore", category=RuntimeWarning)

array = utils.rg.array
rg = utils.rg


def set_ipython_env(
    callbacks: Optional[List[Callback]] = None, interactive: bool = False
):
    logger = logging.getLogger("regelum")
    logger.setLevel(logging.INFO)
    from dataclasses import dataclass

    @dataclass
    class CallbackContainer:
        callbacks: list

    @dataclass
    class ArgvContainer:
        interactive: list
        parallel: bool

    RegelumBase._metadata = {
        "logger": logger,
        "argv": ArgvContainer(interactive, False),
    }
    callbacks = [callback() for callback in callbacks]
    RegelumBase._metadata = {
        "logger": logger,
        "main": CallbackContainer(callbacks),
    "argv": ArgvContainer(interactive, False),
    }
    regelum.main.is_clear_matplotlib_cache_in_callbacks = True
    return callbacks



import copy

from cues import constants, cursor, utils, Survey


class RegelumSurvey(Survey):
    def _draw(self):
        """Print the prompt to console and set user's response."""
        cursor.write(self._init_fmt.format(msg=self._message), color=True)

        if self._legend:
            # If there are only two elems in self._legend:
            if self._header_fmt:
                cursor.write(
                    self._header_fmt.format(
                        *self._legend, space=self._total_legend_fmt_len * " "
                    ),
                    color=True,
                    newlines=1,
                )
                cursor.write(
                    self._legend_fmt.format(*self._scale, space=self._space_btwn),
                    color=True,
                    newlines=3,
                )
            # else, if lengths of _legend and _scale are equal:
            else:
                for pt, desc in zip(self._scale, self._legend):
                    cursor.write(
                        self._legend_fmt.format(val=pt, legend=desc), color=True
                    )
                cursor.write("", newlines=2)

        # For keeping track of location:
        scale_len = len(self._scale)
        max_fields = len(self._fields) * 4
        current_field = max_fields

        # Creates line between survey points:
        deque_scale = self.create_deque(self._scale)
        current_deque_scale = copy.copy(deque_scale)

        center_pt = utils.get_half(scale_len)
        pts = [constants.SURVEY_PT for _ in range(scale_len)]
        pts[center_pt - 1] = constants.SURVEY_PT_FILL
        deque_pts = self.create_deque(pts)
        current_deque_pts = copy.copy(deque_pts)

        min_space_btwn_lines = 5
        max_line_len = max(len(elem) for elem in deque_scale) + min_space_btwn_lines
        line = constants.SURVEY_LINE * max_line_len

        scale_str = ""
        for val in deque_scale:
            scale_str += self._scale_fmt.format(val, length=max_line_len + 1)
        scale_str += "\n"

        messages = [field["message"] for field in self._fields]

        default_margin = 2

        for c, message in enumerate(messages, 1):
            cursor.write(self._msg_fmt.format(count=c, msg=message))

            # Adds space in front:
            margin = " " * (default_margin + utils.get_num_digits(c))

            cursor.write(
                margin + self._pt_fmt.format(*deque_pts, line=line), color=True
            )
            cursor.write(margin + scale_str + "\n")

        cursor.move(y=current_field)

        right = self.keys.get("right")
        left = self.keys.get("left")
        enter = self.keys.get("enter")

        horziontal_num = center_pt
        current_val = 0

        responses = {}

        # Actual drawing:
        while True:
            if current_val == 1:
                break
            cursor.write(
                self._msg_fmt.format(count=current_val + 1, msg=messages[current_val])
            )

            # Adds space in front:
            margin = " " * (default_margin + utils.get_num_digits(current_val + 1))

            cursor.write(
                margin + self._pt_fmt.format(*current_deque_pts, line=line), color=True
            )

            scale_str = ""
            for c, val in enumerate(current_deque_scale, 1):
                temp_line_len = 0
                if c == horziontal_num:
                    val = (
                        "[underline lightslateblue]"
                        + val
                        + "[/underline lightslateblue]"
                    )
                    temp_line_len = max_line_len + len(val)
                scale_str += self._scale_fmt.format(
                    val, length=(temp_line_len or max_line_len + 1)
                )
            scale_str += "\n"
            cursor.write(margin + scale_str + "\n", color=True)

            cursor.move(y=-(current_field - 4))

            key = self.listen_for_key()

            if key == right:
                # If cursor is at very right:
                if current_deque_pts[-1] == constants.SURVEY_PT_FILL:
                    pass
                else:
                    current_deque_pts.appendleft(constants.SURVEY_PT)
                    horziontal_num += 1

            elif key == left:
                # If cursor is at very left:
                if current_deque_pts[0] == constants.SURVEY_PT_FILL:
                    pass
                else:
                    current_deque_pts.append(constants.SURVEY_PT)
                    horziontal_num -= 1

            elif key == enter:
                # Add current scale value to dict
                responses.update(
                    {self._fields[current_val]["name"]: self._scale[horziontal_num - 1]}
                )
                current_val += 1

                # If at the end of the survey, then quit:
                if current_val == scale_len - 1:
                    if self._header_fmt:
                        cursor.clear(max_fields + 4)
                    else:
                        cursor.clear(
                            max_fields + (len(self._legend) + 2 if self._legend else 0)
                        )

                    break
                else:
                    current_field -= 4
                    # Resets values:
                    current_deque_pts = copy.copy(deque_pts)
                    current_deque_scale = copy.copy(deque_scale)
                    horziontal_num = center_pt

            # Resets cursor at top:
            cursor.move(y=current_field)

        self.answer = {self._name: responses}



"""Interfaces for creating optimizables, i.e. dynamic objects (e.g. systems, scenarios) that posses a parameter that can be varied in order to minimize an objective."""

from .optimizers import *
from .core.configs import *
from .core.entities import *
from .core.hooks import *
